"""Purger for RBAC field-scoped entity delete operations."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Collection
from dataclasses import dataclass
from typing import Any, cast

import sqlalchemy as sa
from sqlalchemy.engine import CursorResult
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.manager.data.permission.id import FieldRef
from ai.backend.manager.data.permission.types import FieldType
from ai.backend.manager.errors.repository import UnsupportedCompositePrimaryKeyError
from ai.backend.manager.models.base import Base
from ai.backend.manager.models.rbac_models.entity_field import EntityFieldRow
from ai.backend.manager.repositories.base.purger import BatchPurgerSpec as BaseBatchPurgerSpec
from ai.backend.manager.repositories.base.purger import Purger as BasePurger
from ai.backend.manager.repositories.base.purger import PurgerResult as BasePurgerResult
from ai.backend.manager.repositories.base.purger import TRow

# =============================================================================
# Spec Classes (Batch Only)
# =============================================================================


class RBACFieldBatchPurgerSpec(BaseBatchPurgerSpec[TRow], ABC):
    """Spec for RBAC field batch purge operations.

    Inherits build_subquery() from BaseBatchPurgerSpec.
    Implementations must provide:
    - field_type(): Returns the FieldType for constructing FieldRefs from row PKs
    """

    @abstractmethod
    def field_type(self) -> FieldType:
        """Return the field type for constructing FieldRefs from row primary keys."""
        raise NotImplementedError


# =============================================================================
# Purger Classes
# =============================================================================


@dataclass
class RBACFieldPurger(BasePurger[TRow]):
    """Single-row RBAC field purger by primary key.

    Inherits row_class and pk_value from BasePurger.

    Attributes:
        field_type: The field type for RBAC cleanup.
        field_id: The field ID to delete.
    """

    field_type: FieldType
    field_id: str


@dataclass
class RBACFieldPurgerResult(BasePurgerResult[TRow]):
    """Result of executing a single-row RBAC field purge."""

    pass


@dataclass
class RBACFieldBatchPurger[TRow: Base]:
    """Batch purger for RBAC field-scoped entities.

    Attributes:
        spec: RBACFieldBatchPurgerSpec implementation defining what to delete.
        batch_size: Batch size for chunked deletion (default: 1000).
    """

    spec: RBACFieldBatchPurgerSpec[TRow]
    batch_size: int = 1000


@dataclass
class RBACFieldBatchPurgerResult:
    """Result of RBAC field batch purge operation."""

    deleted_count: int
    deleted_entity_field_count: int


# =============================================================================
# Deletion Helpers (Single Field)
# =============================================================================


async def _delete_entity_field(
    db_sess: SASession,
    field: FieldRef,
) -> None:
    """Delete EntityFieldRow for the given field."""
    await db_sess.execute(
        sa.delete(EntityFieldRow).where(
            sa.and_(
                EntityFieldRow.field_id == field.field_id,
                EntityFieldRow.field_type == field.field_type,
            )
        )
    )


# =============================================================================
# Deletion Helpers (Batch) - Return counts
# =============================================================================


async def _batch_delete_entity_fields(
    db_sess: SASession,
    field_ids: Collection[FieldRef],
) -> int:
    """Delete EntityFieldRows for multiple fields. Returns count of deleted rows."""
    if not field_ids:
        return 0

    conditions = [
        sa.and_(
            EntityFieldRow.field_id == fid.field_id,
            EntityFieldRow.field_type == fid.field_type,
        )
        for fid in field_ids
    ]

    result = await db_sess.execute(sa.delete(EntityFieldRow).where(sa.or_(*conditions)))
    return cast(CursorResult[Any], result).rowcount or 0


# =============================================================================
# Orchestration
# =============================================================================


async def _delete_row_by_pk_returning(
    db_sess: SASession,
    purger: RBACFieldPurger[TRow],
) -> TRow | None:
    """Delete a row by primary key and return the deleted row data."""
    row_class = purger.row_class
    table = row_class.__table__
    pk_columns = list(table.primary_key.columns)

    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Purger only supports single-column primary keys (table: {table.name})",
        )

    stmt = sa.delete(table).where(pk_columns[0] == purger.pk_value).returning(*table.columns)
    result = await db_sess.execute(stmt)
    row_data = result.fetchone()

    if row_data is None:
        return None

    return cast(TRow, row_class(**dict(row_data._mapping)))


# =============================================================================
# Public API
# =============================================================================


async def execute_rbac_field_purger(
    db_sess: SASession,
    purger: RBACFieldPurger[TRow],
) -> RBACFieldPurgerResult[TRow] | None:
    """
    Execute DELETE for a single field-scoped entity by primary key, along with related RBAC entries.

    Operations performed:
    1. Build FieldRef from purger params
    2. Delete EntityFieldRow (field-entity mapping)
    3. Delete the main object row with RETURNING

    Args:
        db_sess: Async SQLAlchemy session (must be writable)
        purger: Purger containing row_class, pk_value, field_type, and field_id

    Returns:
        RBACFieldPurgerResult containing the deleted row, or None if no row matched
    """
    # 1. Build FieldRef from flat params
    field_ref = FieldRef(field_type=purger.field_type, field_id=purger.field_id)

    # 2. Delete RBAC entries (EntityFieldRow)
    await _delete_entity_field(db_sess, field_ref)

    # 3. Delete main row with RETURNING
    row = await _delete_row_by_pk_returning(db_sess, purger)
    if row is None:
        return None

    return RBACFieldPurgerResult(row=row)


async def execute_rbac_field_batch_purger(
    db_sess: SASession,
    purger: RBACFieldBatchPurger[TRow],
) -> RBACFieldBatchPurgerResult:
    """
    Execute batch DELETE for field-scoped entities with RBAC cleanup.

    Deletes rows in batches, cleaning up related RBAC entries for each batch:
    - EntityFieldRows for field_ids

    Args:
        db_sess: Async SQLAlchemy session (must be writable)
        purger: BatchPurger containing spec and batch configuration

    Returns:
        RBACFieldBatchPurgerResult with counts of deleted rows
    """
    total_deleted = 0
    total_entity_field = 0

    # Get table and PK info from subquery
    base_subquery = purger.spec.build_subquery()
    table = cast(sa.Table, base_subquery.froms[0])
    pk_columns = list(table.primary_key.columns)

    if len(pk_columns) != 1:
        raise UnsupportedCompositePrimaryKeyError(
            f"Batch purger only supports single-column primary keys (table: {table.name})",
        )

    pk_col = pk_columns[0]
    field_type = purger.spec.field_type()

    while True:
        # 1. DELETE with RETURNING - get PKs and delete in one query
        sub = purger.spec.build_subquery().subquery()
        pk_subquery = sa.select(sub.c[pk_col.key]).limit(purger.batch_size)

        stmt = sa.delete(table).where(pk_col.in_(pk_subquery)).returning(pk_col)
        result = await db_sess.execute(stmt)
        deleted_pks = result.fetchall()

        if not deleted_pks:
            break

        pk_values = [row[0] for row in deleted_pks]
        batch_deleted = len(pk_values)
        total_deleted += batch_deleted

        # 2. Construct field_ids from deleted PKs
        field_ids: list[FieldRef] = [
            FieldRef(field_type=field_type, field_id=str(pk)) for pk in pk_values
        ]

        # 3. Clean up RBAC entries (after main row deletion - no FK constraint)
        total_entity_field += await _batch_delete_entity_fields(db_sess, field_ids)

        if batch_deleted < purger.batch_size:
            break

    return RBACFieldBatchPurgerResult(
        deleted_count=total_deleted,
        deleted_entity_field_count=total_entity_field,
    )



from .base import ErrorLogAction
from .create import CreateErrorLogAction, CreateErrorLogActionResult
from .search import SearchErrorLogsAction, SearchErrorLogsActionResult

__all__ = (
    "ErrorLogAction",
    "CreateErrorLogAction",
    "CreateErrorLogActionResult",
    "SearchErrorLogsAction",
    "SearchErrorLogsActionResult",
)



"""User GraphQL input types for mutations."""

from __future__ import annotations

from uuid import UUID

import strawberry

from .enums import UserRoleEnumGQL, UserStatusEnumGQL

# Create User Inputs


@strawberry.input(
    name="CreateUserV2Input",
    description=(
        "Added in 26.2.0. Input for creating a new user. "
        "Required fields: email, username, password, domain_name, need_password_change, status, role."
    ),
)
class CreateUserInputGQL:
    """Input for creating a single user."""

    email: str = strawberry.field(
        description="User's email address. Must be unique across the system."
    )
    username: str = strawberry.field(description="Unique username for login.")
    password: str = strawberry.field(description="Initial password for the user.")
    domain_name: str = strawberry.field(description="Domain to assign the user to.")
    need_password_change: bool = strawberry.field(
        description="If true, user must change password on first login."
    )
    status: UserStatusEnumGQL = strawberry.field(description="Initial account status.")
    role: UserRoleEnumGQL = strawberry.field(
        description="User role determining access permissions."
    )
    full_name: str | None = strawberry.field(
        default=None,
        description="User's full display name.",
    )
    description: str | None = strawberry.field(
        default=None,
        description="Optional description or notes about the user.",
    )
    group_ids: list[UUID] | None = strawberry.field(
        default=None,
        description="List of project (group) IDs to assign the user to.",
    )
    allowed_client_ip: list[str] | None = strawberry.field(
        default=None,
        description="Allowed client IP addresses or CIDR ranges.",
    )
    totp_activated: bool = strawberry.field(
        default=False,
        description="Whether to enable TOTP two-factor authentication.",
    )
    resource_policy: str = strawberry.field(
        default="default",
        description="Name of the user resource policy to apply.",
    )
    sudo_session_enabled: bool = strawberry.field(
        default=False,
        description="Whether this user can create sudo sessions.",
    )
    container_uid: int | None = strawberry.field(
        default=None,
        description="User ID (UID) for container processes.",
    )
    container_main_gid: int | None = strawberry.field(
        default=None,
        description="Primary group ID (GID) for container processes.",
    )
    container_gids: list[int] | None = strawberry.field(
        default=None,
        description="Supplementary group IDs for container processes.",
    )


@strawberry.input(
    name="BulkCreateUserV2Input",
    description=(
        "Added in 26.2.0. Input for bulk creating multiple users. "
        "Each user has individual specifications."
    ),
)
class BulkCreateUserInputGQL:
    """Input for bulk creating users with individual specs."""

    users: list[CreateUserInputGQL] = strawberry.field(description="List of user creation inputs.")


# Update User Inputs


@strawberry.input(
    name="UpdateUserV2Input",
    description=(
        "Added in 26.2.0. Input for updating user information. "
        "All fields are optional - only provided fields will be updated."
    ),
)
class UpdateUserInputGQL:
    """Input for updating user information. All fields optional."""

    username: str | None = strawberry.field(
        default=None,
        description="New username.",
    )
    password: str | None = strawberry.field(
        default=None,
        description="New password.",
    )
    full_name: str | None = strawberry.field(
        default=None,
        description="New full display name.",
    )
    description: str | None = strawberry.field(
        default=None,
        description="New description.",
    )
    status: UserStatusEnumGQL | None = strawberry.field(
        default=None,
        description="New account status.",
    )
    role: UserRoleEnumGQL | None = strawberry.field(
        default=None,
        description="New user role.",
    )
    domain_name: str | None = strawberry.field(
        default=None,
        description="New domain assignment.",
    )
    group_ids: list[UUID] | None = strawberry.field(
        default=None,
        description="New project (group) assignments. Replaces existing assignments.",
    )
    allowed_client_ip: list[str] | None = strawberry.field(
        default=None,
        description="New allowed client IP addresses or CIDR ranges.",
    )
    need_password_change: bool | None = strawberry.field(
        default=None,
        description="Set password change requirement.",
    )
    resource_policy: str | None = strawberry.field(
        default=None,
        description="New user resource policy name.",
    )
    sudo_session_enabled: bool | None = strawberry.field(
        default=None,
        description="Enable or disable sudo session capability.",
    )
    main_access_key: str | None = strawberry.field(
        default=None,
        description="Set the primary API access key.",
    )
    container_uid: int | None = strawberry.field(
        default=None,
        description="New container user ID.",
    )
    container_main_gid: int | None = strawberry.field(
        default=None,
        description="New container primary group ID.",
    )
    container_gids: list[int] | None = strawberry.field(
        default=None,
        description="New container supplementary group IDs.",
    )


# Delete User Inputs


@strawberry.input(
    name="DeleteUsersV2Input",
    description=(
        "Added in 26.2.0. Input for soft-deleting multiple users. "
        "Soft delete changes user status to DELETED but preserves data."
    ),
)
class DeleteUsersInputGQL:
    """Input for soft-deleting multiple users."""

    user_ids: list[UUID] = strawberry.field(description="List of user UUIDs to soft-delete.")


# Purge User Inputs


@strawberry.input(
    name="PurgeUserV2Input",
    description=(
        "Added in 26.2.0. Input for permanently deleting a user and all associated data. "
        "This action is irreversible."
    ),
)
class PurgeUserInputGQL:
    """Input for permanently deleting a single user."""

    user_id: UUID = strawberry.field(description="UUID of the user to purge.")


@strawberry.input(
    name="PurgeUsersV2Input",
    description=(
        "Added in 26.2.0. Input for permanently deleting multiple users. "
        "This action is irreversible."
    ),
)
class PurgeUsersInputGQL:
    """Input for permanently deleting multiple users."""

    user_ids: list[UUID] = strawberry.field(description="List of user UUIDs to purge.")



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import StrEnum
from typing import TYPE_CHECKING, Any

from ai.backend.common.types import AgentSelectionStrategy, ResourceSlot, SessionTypes

if TYPE_CHECKING:
    from ai.backend.manager.models.scaling_group.types import FairShareScalingGroupSpec


class SchedulerType(StrEnum):
    """Scheduler type for session scheduling."""

    FIFO = "fifo"
    LIFO = "lifo"
    DRF = "drf"
    FAIR_SHARE = "fair-share"


@dataclass
class ScalingGroupStatus:
    """Status information for a scaling group."""

    is_active: bool
    is_public: bool


@dataclass
class ScalingGroupMetadata:
    """Metadata for a scaling group."""

    description: str
    created_at: datetime


@dataclass
class ScalingGroupNetworkConfig:
    """Network configuration for a scaling group."""

    wsproxy_addr: str
    wsproxy_api_token: str
    use_host_network: bool


@dataclass
class ScalingGroupDriverConfig:
    """Driver configuration for a scaling group."""

    name: str
    options: Mapping[str, Any]


@dataclass
class ScalingGroupSchedulerOptions:
    """Scheduler options for a scaling group."""

    allowed_session_types: list[SessionTypes]
    pending_timeout: timedelta
    config: Mapping[str, Any]
    agent_selection_strategy: AgentSelectionStrategy
    agent_selector_config: Mapping[str, Any]
    enforce_spreading_endpoint_replica: bool
    allow_fractional_resource_fragmentation: bool
    route_cleanup_target_statuses: list[str]

    def to_json(self) -> dict[str, Any]:
        """Convert scheduler options to JSON-serializable dict."""
        return {
            "allowed_session_types": [st.value for st in self.allowed_session_types],
            "pending_timeout": self.pending_timeout.total_seconds(),
            "config": dict(self.config),
            "agent_selection_strategy": self.agent_selection_strategy.value,
            "agent_selector_config": dict(self.agent_selector_config),
            "enforce_spreading_endpoint_replica": self.enforce_spreading_endpoint_replica,
            "allow_fractional_resource_fragmentation": self.allow_fractional_resource_fragmentation,
            "route_cleanup_target_statuses": self.route_cleanup_target_statuses,
        }


@dataclass
class ScalingGroupSchedulerConfig:
    """Scheduler configuration for a scaling group."""

    name: SchedulerType
    options: ScalingGroupSchedulerOptions


@dataclass
class ScalingGroupData:
    name: str
    status: ScalingGroupStatus
    metadata: ScalingGroupMetadata
    network: ScalingGroupNetworkConfig
    driver: ScalingGroupDriverConfig
    scheduler: ScalingGroupSchedulerConfig
    fair_share_spec: FairShareScalingGroupSpec


@dataclass
class ScalingGroupListResult:
    """Result of searching scaling groups."""

    items: list[ScalingGroupData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass(frozen=True)
class ResourceInfo:
    """Resource information for a scaling group.

    Provides aggregated resource metrics:
    - capacity: Sum of available_slots from ALIVE, schedulable agents
    - used: Sum of occupied_slots from kernels in RUNNING/TERMINATING status
    - free: capacity - used
    """

    capacity: ResourceSlot
    """Total available resources from ALIVE, schedulable agents."""

    used: ResourceSlot
    """Currently occupied resources from active kernels."""

    free: ResourceSlot
    """Available resources (capacity - used)."""



"""Auto-scaling rule data types for deployment repository."""

import uuid
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from ai.backend.common.types import (
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
)


@dataclass
class AutoScalingRuleData:
    """Data structure for auto-scaling rule."""

    rule_id: uuid.UUID
    endpoint_id: uuid.UUID
    metric_source: AutoScalingMetricSource
    metric_name: str
    threshold: Decimal
    comparator: AutoScalingMetricComparator
    step_size: int
    cooldown_seconds: int
    min_replicas: int | None = None
    max_replicas: int | None = None
    enabled: bool = True
    created_at: datetime | None = None
    updated_at: datetime | None = None
    last_triggered_at: datetime | None = None



"""add_degraded_route_status

Revision ID: dd15fd17f232
Revises: 4cfe35b1e60a
Create Date: 2025-11-20 15:13:46.120822

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "dd15fd17f232"
down_revision = "4cfe35b1e60a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Add 'degraded' value to routestatus enum
    op.execute("ALTER TYPE routestatus ADD VALUE IF NOT EXISTS 'degraded'")


def downgrade() -> None:
    # Note: PostgreSQL does not support removing enum values
    # Manual intervention required if downgrade is needed
    pass



"""Post-processor for marking next schedule type."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.scheduler.types import ScheduleType

from .base import PostProcessor, PostProcessorContext

if TYPE_CHECKING:
    from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

log = BraceStyleAdapter(logging.getLogger(__name__))


# Mapping from target session status to next schedule type.
# After a session transitions to a status, the coordinator marks the corresponding
# schedule type to trigger the next phase.
STATUS_TO_NEXT_SCHEDULE_TYPE: dict[SessionStatus, ScheduleType] = {
    SessionStatus.PENDING: ScheduleType.SCHEDULE,
    SessionStatus.SCHEDULED: ScheduleType.CHECK_PRECONDITION,
    SessionStatus.PREPARED: ScheduleType.START,
    SessionStatus.DEPRIORITIZING: ScheduleType.DEPRIORITIZE,
    SessionStatus.TERMINATING: ScheduleType.TERMINATE,
}


class ScheduleMarkingPostProcessor(PostProcessor):
    """Post-processor that marks the next schedule type based on target statuses."""

    def __init__(self, scheduling_controller: SchedulingController) -> None:
        self._scheduling_controller = scheduling_controller

    async def execute(self, context: PostProcessorContext) -> None:
        """Mark the next schedule types for all target statuses."""
        if not context.target_statuses:
            return

        # Collect unique schedule types to mark (deduplicate since multiple statuses
        # might map to the same schedule type)
        schedule_types_to_mark: set[ScheduleType] = set()
        for target_status in context.target_statuses:
            next_schedule_type = STATUS_TO_NEXT_SCHEDULE_TYPE.get(target_status)
            if next_schedule_type is not None:
                schedule_types_to_mark.add(next_schedule_type)

        if not schedule_types_to_mark:
            return

        # Mark all schedule types in batch
        await self._scheduling_controller.mark_scheduling_needed(list(schedule_types_to_mark))
        log.debug(
            "Marked {} schedule type(s) for status transitions to {}",
            len(schedule_types_to_mark),
            context.target_statuses,
        )



"""Project V2 GraphQL scope types."""

from __future__ import annotations

import strawberry


@strawberry.input(
    name="DomainProjectV2Scope",
    description=(
        "Added in 26.2.0. Scope for querying projects within a specific domain. "
        "Used to restrict project queries to a particular domain context."
    ),
)
class DomainProjectScope:
    """Scope for domain-level project queries."""

    domain_name: str = strawberry.field(
        description="Domain name to scope the query. Only projects belonging to this domain will be returned."
    )



"""add-keypair-resource-policy

Revision ID: dbc1e053b880
Revises: 2b0931e4a059
Create Date: 2019-02-07 15:30:54.861821

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import text

from ai.backend.common.types import DefaultForUnspecified

# revision identifiers, used by Alembic.
revision = "dbc1e053b880"
down_revision = "2b0931e4a059"
branch_labels = None
depends_on = None


default_for_unspecified_choices = [*map(str, DefaultForUnspecified)]
default_for_unspecified = postgresql.ENUM(
    *default_for_unspecified_choices,
    name="default_for_unspecified",
)


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "keypair_resource_policies",
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column(
            "default_for_unspecified",
            type_=sa.Enum(*default_for_unspecified_choices, name="default_for_unspecified"),
            nullable=False,
        ),
        sa.Column("total_resource_slots", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("max_concurrent_sessions", sa.Integer(), nullable=False),
        sa.Column("max_containers_per_session", sa.Integer(), nullable=False),
        sa.Column("max_vfolder_count", sa.Integer(), nullable=False),
        sa.Column("max_vfolder_size", sa.BigInteger(), nullable=False),
        sa.Column("allowed_vfolder_hosts", postgresql.ARRAY(sa.String()), nullable=False),
        sa.PrimaryKeyConstraint("name", name=op.f("pk_keypair_resource_policies")),
    )
    # Create a default resource policy
    query = """
    INSERT INTO keypair_resource_policies
    VALUES (
        'default',
        now(),
        'UNLIMITED',
        '{}'::jsonb,
        30,
        1,
        10,
        0,
        array['local']
    );
    """
    connection = op.get_bind()
    connection.execute(text(query))
    query = """
    UPDATE keypairs
    SET resource_policy = 'default';
    """
    connection.execute(text(query))
    print("\n!!! NOTICE !!!\n")
    print("Created a default resource policy and linked all keypairs to it.")
    print("Please inspect and adjust it!\n")
    op.alter_column(
        "keypairs",
        "resource_policy",
        existing_type=sa.VARCHAR(),
        type_=sa.String(length=256),
        nullable=False,
    )
    op.create_foreign_key(
        op.f("fk_keypairs_resource_policy_keypair_resource_policies"),
        "keypairs",
        "keypair_resource_policies",
        ["resource_policy"],
        ["name"],
    )
    op.drop_column("keypairs", "concurrency_limit")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        op.f("fk_keypairs_resource_policy_keypair_resource_policies"),
        "keypairs",
        type_="foreignkey",
    )
    op.add_column(
        "keypairs", sa.Column("concurrency_limit", sa.INTEGER(), autoincrement=False, nullable=True)
    )
    op.alter_column(
        "keypairs",
        "resource_policy",
        existing_type=sa.String(length=256),
        type_=sa.VARCHAR(),
        nullable=True,
    )
    op.drop_table("keypair_resource_policies")
    default_for_unspecified.drop(op.get_bind())
    # ### end Alembic commands ###



"""ProjectV2 GraphQL API package.

Added in 26.2.0. Provides structured project management API with typed fields
replacing JSON scalars and organized into logical field groups.
"""

from .resolver import (
    admin_projects_v2,
    domain_projects_v2,
    project_domain_v2,
    project_v2,
)
from .types import (
    ProjectBasicInfoGQL,
    ProjectLifecycleInfoGQL,
    ProjectOrganizationInfoGQL,
    ProjectStorageInfoGQL,
    ProjectTypeEnum,
    ProjectV2Connection,
    ProjectV2Edge,
    ProjectV2GQL,
    VFolderHostPermissionEntryGQL,
    VFolderHostPermissionEnum,
)

__all__ = [
    # Queries
    "admin_projects_v2",
    "domain_projects_v2",
    "project_domain_v2",
    "project_v2",
    # Enums
    "ProjectTypeEnum",
    "VFolderHostPermissionEnum",
    # Nested types
    "ProjectBasicInfoGQL",
    "ProjectOrganizationInfoGQL",
    "VFolderHostPermissionEntryGQL",
    "ProjectStorageInfoGQL",
    "ProjectLifecycleInfoGQL",
    # Node types
    "ProjectV2GQL",
    "ProjectV2Edge",
    "ProjectV2Connection",
]



from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseAction, BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.object_permission import ObjectPermissionData
from ai.backend.manager.models.rbac_models.permission.object_permission import ObjectPermissionRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.purger import Purger


@dataclass
class ObjectPermissionAction(BaseAction):
    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.OBJECT_PERMISSION


@dataclass
class CreateObjectPermissionAction(ObjectPermissionAction):
    creator: Creator[ObjectPermissionRow]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.OBJECT_PERMISSION

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateObjectPermissionActionResult(BaseActionResult):
    data: ObjectPermissionData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id) if self.data else None


@dataclass
class DeleteObjectPermissionAction(ObjectPermissionAction):
    purger: Purger[ObjectPermissionRow]

    @override
    def entity_id(self) -> str | None:
        return str(self.purger.pk_value)

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.OBJECT_PERMISSION

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeleteObjectPermissionActionResult(BaseActionResult):
    data: ObjectPermissionData | None

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id) if self.data else None



import uuid
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Any, cast

import sqlalchemy as sa
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import contains_eager, selectinload

from ai.backend.manager.data.permission.entity import EntityData, EntityListResult
from ai.backend.manager.data.permission.id import ObjectId, ScopeId
from ai.backend.manager.data.permission.object_permission import (
    ObjectPermissionCreateInputBeforeRoleCreation,
    ObjectPermissionListResult,
)
from ai.backend.manager.data.permission.permission import (
    PermissionListResult,
)
from ai.backend.manager.data.permission.role import (
    AssignedUserData,
    AssignedUserListResult,
    RoleListResult,
    RolePermissionsUpdateInput,
    UserRoleAssignmentInput,
    UserRoleRevocationInput,
)
from ai.backend.manager.data.permission.status import (
    RoleStatus,
)
from ai.backend.manager.data.permission.types import (
    OperationType,
    ScopeData,
    ScopeListResult,
    ScopeType,
)
from ai.backend.manager.errors.common import ObjectNotFound
from ai.backend.manager.errors.permission import RoleAlreadyAssigned, RoleNotAssigned, RoleNotFound
from ai.backend.manager.models.domain.row import DomainRow
from ai.backend.manager.models.group.row import GroupRow
from ai.backend.manager.models.rbac_models.association_scopes_entities import (
    AssociationScopesEntitiesRow,
)
from ai.backend.manager.models.rbac_models.permission.object_permission import ObjectPermissionRow
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow
from ai.backend.manager.models.rbac_models.role import RoleRow
from ai.backend.manager.models.rbac_models.user_role import UserRoleRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.base.creator import Creator, execute_creator
from ai.backend.manager.repositories.base.purger import Purger, execute_purger
from ai.backend.manager.repositories.base.querier import BatchQuerier, execute_batch_querier
from ai.backend.manager.repositories.base.updater import Updater, execute_updater
from ai.backend.manager.repositories.permission_controller.creators import (
    ObjectPermissionCreatorSpec,
    PermissionCreatorSpec,
)
from ai.backend.manager.repositories.permission_controller.types import (
    ObjectPermissionSearchScope,
    PermissionSearchScope,
)


@dataclass
class CreateRoleInput:
    """Input for creating a role with object permissions."""

    creator: Creator[RoleRow]
    object_permissions: Sequence[ObjectPermissionCreateInputBeforeRoleCreation]


class PermissionDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def create_role(self, input_data: CreateRoleInput) -> RoleRow:
        """
        Create a new role with object permissions.

        All related entities are created in a single transaction.

        Args:
            input_data: Input containing creator and object permissions

        Returns:
            Created role row
        """
        async with self._db.begin_session() as db_session:
            # 1. Create role
            result = await execute_creator(db_session, input_data.creator)
            role_row = result.row

            await db_session.refresh(role_row)
            return role_row

    async def create_permission(
        self,
        creator: Creator[PermissionRow],
    ) -> PermissionRow:
        """
        Create a permission.

        Args:
            creator: Permission creator defining the permission to create

        Returns:
            Created permission row
        """
        async with self._db.begin_session() as db_session:
            perm_row = await self._add_permission_to_group(db_session, creator)
            await db_session.refresh(perm_row)
            return perm_row

    async def create_object_permission(
        self,
        creator: Creator[ObjectPermissionRow],
    ) -> ObjectPermissionRow:
        """
        Create an object permission for a role.

        Args:
            creator: Object permission creator defining the permission to create

        Returns:
            Created object permission row
        """
        async with self._db.begin_session() as db_session:
            obj_perm_row = await self._add_object_permission_to_role(db_session, creator)
            await db_session.refresh(obj_perm_row)
            return obj_perm_row

    async def delete_permission(
        self,
        purger: Purger[PermissionRow],
    ) -> PermissionRow:
        """
        Delete a permission.

        Args:
            purger: Purger with permission ID

        Returns:
            Deleted permission row

        Raises:
            ObjectNotFound: If permission does not exist
        """
        async with self._db.begin_session() as db_session:
            result = await execute_purger(db_session, purger)
            if result is None:
                raise ObjectNotFound(f"Permission with ID {purger.pk_value} does not exist.")
            return result.row

    async def delete_object_permission(
        self,
        purger: Purger[ObjectPermissionRow],
    ) -> ObjectPermissionRow | None:
        """
        Delete an object permission.

        Args:
            purger: Purger with object permission ID

        Returns:
            Deleted object permission row, or None if not found
        """
        async with self._db.begin_session() as db_session:
            result = await execute_purger(db_session, purger)
            return result.row if result else None

    async def _get_role(self, db_session: SASession, role_id: uuid.UUID) -> RoleRow:
        stmt = sa.select(RoleRow).where(RoleRow.id == role_id)
        role_row = await db_session.scalar(stmt)
        result = role_row
        if result is None:
            raise ObjectNotFound(f"Role with ID {role_id} does not exist.")
        return result

    # ============================================================
    # Private Helper Functions (for use within transactions)
    # ============================================================

    async def _add_permission_to_group(
        self,
        db_session: SASession,
        creator: Creator[PermissionRow],
    ) -> PermissionRow:
        """Add a permission (private, within transaction)."""
        result = await execute_creator(db_session, creator)
        return result.row

    async def _add_object_permission_to_role(
        self,
        db_session: SASession,
        creator: Creator[ObjectPermissionRow],
    ) -> ObjectPermissionRow:
        """Add an object permission to a role (private, within transaction)."""
        result = await execute_creator(db_session, creator)
        return result.row

    async def _remove_permission(
        self,
        db_session: SASession,
        purger: Purger[PermissionRow],
    ) -> None:
        """Remove a permission (private, within transaction)."""
        await execute_purger(db_session, purger)

    async def _remove_object_permission_from_role(
        self,
        db_session: SASession,
        purger: Purger[ObjectPermissionRow],
    ) -> None:
        """Remove an object permission from a role (private, within transaction)."""
        await execute_purger(db_session, purger)

    async def update_role(self, updater: Updater[RoleRow]) -> RoleRow:
        async with self._db.begin_session() as db_session:
            result = await execute_updater(db_session, updater)
            if result is None:
                raise ObjectNotFound(f"Role with ID {updater.pk_value} does not exist.")
            return result.row

    async def delete_role(self, updater: Updater[RoleRow]) -> RoleRow:
        async with self._db.begin_session() as db_session:
            result = await execute_updater(db_session, updater)
            if result is None:
                raise ObjectNotFound(f"Role with ID {updater.pk_value} does not exist.")
            return result.row

    async def purge_role(self, purger: Purger[RoleRow]) -> RoleRow:
        async with self._db.begin_session() as db_session:
            result = await execute_purger(db_session, purger)
            if result is None:
                raise ObjectNotFound(f"Role with ID {purger.pk_value} does not exist.")
            return result.row

    async def assign_role(self, data: UserRoleAssignmentInput) -> UserRoleRow:
        async with self._db.begin_session() as db_session:
            user_role_row = UserRoleRow.from_input(data)
            try:
                db_session.add(user_role_row)
                await db_session.flush()
                await db_session.refresh(user_role_row)
                return user_role_row
            except IntegrityError as e:
                raise RoleAlreadyAssigned(
                    f"Role {data.role_id} is already assigned to user {data.user_id}."
                ) from e

    async def revoke_role(self, data: UserRoleRevocationInput) -> uuid.UUID:
        async with self._db.begin_session() as db_session:
            stmt = (
                sa.select(UserRoleRow)
                .where(UserRoleRow.user_id == data.user_id)
                .where(UserRoleRow.role_id == data.role_id)
            )
            user_role_row = await db_session.scalar(stmt)

            if user_role_row is None:
                raise RoleNotAssigned(
                    f"Role {data.role_id} is not assigned to user {data.user_id}."
                )

            user_role_id = user_role_row.id
            await db_session.delete(user_role_row)
            await db_session.flush()
            return user_role_id

    async def update_role_permissions(
        self,
        input_data: RolePermissionsUpdateInput,
    ) -> RoleRow:
        """
        Update role permissions in batch.

        Args:
            input_data: Batch update input containing scoped and object permissions

        Returns:
            Updated role with refreshed relationships

        Raises:
            ObjectNotFound: If role does not exist
        """
        async with self._db.begin_session() as db_session:
            # 0. Verify role exists
            role_row = await self._get_role(db_session, input_data.role_id)

            # 1. Add scoped permissions
            for scoped_perm_input in input_data.add_scoped_permissions:
                perm_creator = Creator(
                    spec=PermissionCreatorSpec(
                        role_id=input_data.role_id,
                        scope_type=scoped_perm_input.scope_type,
                        scope_id=scoped_perm_input.scope_id,
                        entity_type=scoped_perm_input.entity_type,
                        operation=scoped_perm_input.operation,
                    )
                )
                await self._add_permission_to_group(db_session, perm_creator)

            # 2. Remove scoped permissions
            for perm_id in input_data.remove_scoped_permission_ids:
                perm_purger = Purger(row_class=PermissionRow, pk_value=perm_id)
                await self._remove_permission(db_session, perm_purger)

            # 3. Add object permissions
            for obj_perm_input in input_data.add_object_permissions:
                obj_perm_creator = Creator(
                    spec=ObjectPermissionCreatorSpec(
                        role_id=input_data.role_id,
                        entity_type=obj_perm_input.entity_type,
                        entity_id=obj_perm_input.entity_id,
                        operation=obj_perm_input.operation,
                        status=obj_perm_input.status,
                    )
                )
                await self._add_object_permission_to_role(db_session, obj_perm_creator)

            # 4. Remove object permissions
            for obj_perm_id in input_data.remove_object_permission_ids:
                obj_perm_purger = Purger(row_class=ObjectPermissionRow, pk_value=obj_perm_id)
                await self._remove_object_permission_from_role(db_session, obj_perm_purger)

            # 5. Refresh and return
            await db_session.refresh(role_row)
            return role_row

    async def get_role(self, role_id: uuid.UUID) -> RoleRow | None:
        async with self._db.begin_readonly_session_read_committed() as db_session:
            try:
                result = await self._get_role(db_session, role_id)
            except ObjectNotFound:
                return None
            return result

    async def get_user_roles(self, user_id: uuid.UUID) -> list[RoleRow]:
        async with self._db.begin_readonly_session_read_committed() as db_session:
            j = (
                sa.join(
                    RoleRow,
                    UserRoleRow,
                    RoleRow.id == UserRoleRow.role_id,
                )
                .join(
                    ObjectPermissionRow,
                    RoleRow.id == ObjectPermissionRow.role_id,
                )
                .join(
                    PermissionRow,
                    RoleRow.id == PermissionRow.role_id,
                )
            )
            stmt = (
                sa.select(RoleRow)
                .select_from(j)
                .where(UserRoleRow.user_id == user_id)
                .options(
                    selectinload(RoleRow.object_permission_rows),
                )
            )

            result = await db_session.scalars(stmt)
            return list(result.all())

    async def get_entity_mapped_scopes(
        self, target_object_id: ObjectId
    ) -> list[AssociationScopesEntitiesRow]:
        async with self._db.begin_readonly_session_read_committed() as db_session:
            stmt = sa.select(AssociationScopesEntitiesRow).where(
                AssociationScopesEntitiesRow.entity_id == target_object_id.entity_id,
                AssociationScopesEntitiesRow.entity_type == target_object_id.entity_type.value,
            )
            result = await db_session.scalars(stmt)
            return list(result.all())

    async def check_scope_permission_exist(
        self,
        user_id: uuid.UUID,
        scope_id: ScopeId,
        operation: OperationType,
    ) -> bool:
        role_query = (
            sa.select(sa.func.exist())
            .select_from(
                sa.join(RoleRow, UserRoleRow, RoleRow.id == UserRoleRow.role_id).join(
                    PermissionRow, RoleRow.id == PermissionRow.role_id
                )
            )
            .where(
                sa.and_(
                    RoleRow.status == RoleStatus.ACTIVE,
                    UserRoleRow.user_id == user_id,
                    sa.or_(
                        PermissionRow.scope_type == ScopeType.GLOBAL,
                        PermissionRow.scope_id == scope_id.scope_id,
                    ),
                    PermissionRow.operation == operation,
                )
            )
        )
        async with self._db.begin_readonly_session_read_committed() as db_session:
            result = await db_session.scalar(role_query)
            return result or False

    def _make_query_statement_for_object_permission(
        self,
        user_id: uuid.UUID,
        object_ids: Iterable[ObjectId],
    ) -> sa.sql.Select[Any]:
        object_id_for_cond = [obj_id.entity_id for obj_id in object_ids]
        return (
            sa.select(RoleRow)
            .select_from(
                sa.join(RoleRow, UserRoleRow, RoleRow.id == UserRoleRow.role_id)
                .join(PermissionRow, RoleRow.id == PermissionRow.role_id)
                .join(
                    AssociationScopesEntitiesRow,
                    sa.and_(
                        PermissionRow.scope_id == AssociationScopesEntitiesRow.scope_id,
                        PermissionRow.scope_type == AssociationScopesEntitiesRow.scope_type,
                    ),
                )
                .join(ObjectPermissionRow, RoleRow.id == ObjectPermissionRow.role_id)
            )
            .where(
                sa.and_(
                    RoleRow.status == RoleStatus.ACTIVE,
                    UserRoleRow.user_id == user_id,
                    sa.or_(
                        PermissionRow.scope_type == ScopeType.GLOBAL,
                        AssociationScopesEntitiesRow.entity_id.in_(object_id_for_cond),
                        ObjectPermissionRow.entity_id.in_(object_id_for_cond),
                    ),
                )
            )
            .options(
                contains_eager(RoleRow.object_permission_rows),
            )
        )

    def _make_query_statement_for_object_permissions(
        self,
        user_id: uuid.UUID,
        object_ids: Iterable[ObjectId],
        operation: OperationType,
    ) -> sa.sql.Select[Any]:
        object_id_for_cond = [obj_id.entity_id for obj_id in object_ids]
        return (
            sa.select(RoleRow)
            .select_from(
                sa.join(RoleRow, UserRoleRow, RoleRow.id == UserRoleRow.role_id)
                .join(PermissionRow, RoleRow.id == PermissionRow.role_id)
                .join(
                    AssociationScopesEntitiesRow,
                    sa.and_(
                        PermissionRow.scope_id == AssociationScopesEntitiesRow.scope_id,
                        PermissionRow.scope_type == AssociationScopesEntitiesRow.scope_type,
                    ),
                )
                .join(ObjectPermissionRow, RoleRow.id == ObjectPermissionRow.role_id)
            )
            .where(
                sa.and_(
                    RoleRow.status == RoleStatus.ACTIVE,
                    UserRoleRow.user_id == user_id,
                    sa.or_(
                        sa.and_(
                            PermissionRow.scope_type == ScopeType.GLOBAL,
                            PermissionRow.operation == operation,
                        ),
                        sa.and_(
                            AssociationScopesEntitiesRow.entity_id.in_(object_id_for_cond),
                            PermissionRow.operation == operation,
                        ),
                        sa.and_(
                            ObjectPermissionRow.entity_id.in_(object_id_for_cond),
                            ObjectPermissionRow.operation == operation,
                        ),
                    ),
                )
            )
            .options(
                contains_eager(RoleRow.object_permission_rows),
            )
        )

    async def check_object_permission_exist(
        self,
        user_id: uuid.UUID,
        object_id: ObjectId,
        operation: OperationType,
    ) -> bool:
        role_query = self._make_query_statement_for_object_permissions(
            user_id, [object_id], operation
        )
        async with self._db.begin_readonly_session_read_committed() as db_session:
            result = await db_session.scalars(role_query)
            role_rows = cast(list[RoleRow], result.all())
            return len(role_rows) > 0

    async def check_batch_object_permission_exist(
        self,
        user_id: uuid.UUID,
        object_ids: Iterable[ObjectId],
        operation: OperationType,
    ) -> dict[ObjectId, bool]:
        result: dict[ObjectId, bool] = dict.fromkeys(object_ids, False)
        role_query = self._make_query_statement_for_object_permissions(
            user_id, object_ids, operation
        )
        async with self._db.begin_readonly_session_read_committed() as db_session:
            role_rows_result = await db_session.scalars(role_query)
            role_rows = list(role_rows_result.all())

            for role in role_rows:
                for op in role.object_permission_rows:
                    object_id = op.object_id()
                    result[object_id] = True
        return result

    async def search_roles(
        self,
        querier: BatchQuerier,
    ) -> RoleListResult:
        """Searches roles with pagination and filtering.

        Uses LEFT JOIN with ObjectPermissionRow to support
        entity-based filtering. The JOIN is always performed to
        simplify the implementation, with distinct() used to prevent duplicates.
        """
        async with self._db.begin_readonly_session() as db_sess:
            # Build query with LEFT JOIN to support entity filtering
            query = (
                sa.select(RoleRow)
                .outerjoin(
                    ObjectPermissionRow,
                    RoleRow.id == ObjectPermissionRow.role_id,
                )
                .distinct()
            )

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.RoleRow.to_data() for row in result.rows]

            return RoleListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_permissions(
        self,
        querier: BatchQuerier,
        scope: PermissionSearchScope | None = None,
    ) -> PermissionListResult:
        """Searches permissions with pagination and filtering."""
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            query = sa.select(PermissionRow)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
                scope,
            )

            items = [row.PermissionRow.to_data() for row in result.rows]

            return PermissionListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_object_permissions(
        self,
        querier: BatchQuerier,
        scope: ObjectPermissionSearchScope | None = None,
    ) -> ObjectPermissionListResult:
        """Searches object permissions with pagination and filtering."""
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            query = sa.select(ObjectPermissionRow)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
                scope,
            )

            items = [row.ObjectPermissionRow.to_data() for row in result.rows]

            return ObjectPermissionListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def get_role_with_permissions(self, role_id: uuid.UUID) -> RoleRow:
        """Get role with eagerly loaded permissions only (no users)."""
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            stmt = (
                sa.select(RoleRow)
                .where(RoleRow.id == role_id)
                .options(
                    selectinload(RoleRow.object_permission_rows),
                )
            )
            result = await db_sess.execute(stmt)
            role_row = result.scalar_one_or_none()
            if role_row is None:
                raise RoleNotFound(f"Role with ID {role_id} does not exist.")
            return role_row

    async def search_users_assigned_to_role(
        self,
        querier: BatchQuerier,
    ) -> AssignedUserListResult:
        """Searches users assigned to a specific role with pagination and filtering."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(UserRow, UserRoleRow).select_from(
                sa.join(
                    UserRow,
                    UserRoleRow,
                    UserRoleRow.user_id == UserRow.uuid,
                )
            )
            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [
                AssignedUserData(
                    id=row.UserRoleRow.id,
                    user_id=row.UserRow.uuid,
                    role_id=row.UserRoleRow.role_id,
                    granted_by=row.UserRoleRow.granted_by,
                    granted_at=row.UserRoleRow.granted_at,
                )
                for row in result.rows
            ]

            return AssignedUserListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_domain_scopes(
        self,
        querier: BatchQuerier,
    ) -> ScopeListResult:
        """Search all domains using BatchQuerier."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(DomainRow.name)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [
                ScopeData(
                    id=ScopeId(scope_type=ScopeType.DOMAIN, scope_id=row.name),
                    name=row.name,
                )
                for row in result.rows
            ]

            return ScopeListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_project_scopes(
        self,
        querier: BatchQuerier,
    ) -> ScopeListResult:
        """Search all projects using BatchQuerier."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(GroupRow.id, GroupRow.name)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [
                ScopeData(
                    id=ScopeId(scope_type=ScopeType.PROJECT, scope_id=str(row.id)),
                    name=row.name,
                )
                for row in result.rows
            ]

            return ScopeListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_user_scopes(
        self,
        querier: BatchQuerier,
    ) -> ScopeListResult:
        """Search all users using BatchQuerier."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(UserRow.uuid, UserRow.username, UserRow.email)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [
                ScopeData(
                    id=ScopeId(scope_type=ScopeType.USER, scope_id=str(row.uuid)),
                    name=row.username if row.username is not None else row.email,
                )
                for row in result.rows
            ]

            return ScopeListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_entities_in_scope(
        self,
        querier: BatchQuerier,
    ) -> EntityListResult:
        """Search entities within a scope."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(
                AssociationScopesEntitiesRow.entity_id,
                AssociationScopesEntitiesRow.entity_type,
            )

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [
                EntityData(
                    entity_type=row.entity_type,
                    entity_id=row.entity_id,
                )
                for row in result.rows
            ]

            return EntityListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )



"""
Adapters to convert deployment DTOs to repository BatchQuerier objects.
Handles conversion of filter, order, and pagination parameters.
Also provides data-to-DTO conversion functions.
"""

from __future__ import annotations

from datetime import UTC, datetime
from pathlib import PurePosixPath
from typing import Any
from uuid import UUID, uuid4

from ai.backend.common.data.model_deployment.types import DeploymentStrategy
from ai.backend.common.data.model_deployment.types import (
    RouteStatus as CommonRouteStatus,
)
from ai.backend.common.data.model_deployment.types import (
    RouteTrafficStatus as CommonRouteTrafficStatus,
)
from ai.backend.common.dto.manager.deployment import (
    ClusterConfigDTO,
    CreateDeploymentRequest,
    CreateRevisionRequest,
    DeploymentDTO,
    DeploymentFilter,
    DeploymentOrder,
    ModelMountConfigDTO,
    ModelRuntimeConfigDTO,
    NetworkConfigDTO,
    ReplicaStateDTO,
    ResourceConfigDTO,
    RevisionDTO,
    RevisionFilter,
    RevisionOrder,
    RouteDTO,
    RouteFilter,
    RouteOrder,
    SearchDeploymentsRequest,
    SearchRevisionsRequest,
    SearchRoutesRequest,
)
from ai.backend.common.dto.manager.deployment.types import (
    DeploymentOrderField,
    OrderDirection,
    RevisionOrderField,
    RouteOrderField,
)
from ai.backend.common.types import ClusterMode, RuntimeVariant
from ai.backend.manager.api.adapter import BaseFilterAdapter
from ai.backend.manager.data.deployment.creator import (
    DeploymentPolicyConfig,
    ModelRevisionCreator,
    NewDeploymentCreator,
    VFolderMountsCreator,
)
from ai.backend.manager.data.deployment.types import (
    DeploymentMetadata,
    DeploymentNetworkSpec,
    ExecutionSpec,
    ModelDeploymentData,
    ModelRevisionData,
    MountInfo,
    ReplicaSpec,
    ResourceSpec,
    RouteInfo,
)
from ai.backend.manager.data.deployment.types import (
    RouteStatus as ManagerRouteStatus,
)
from ai.backend.manager.data.deployment.types import (
    RouteTrafficStatus as ManagerRouteTrafficStatus,
)
from ai.backend.manager.models.deployment_policy import BlueGreenSpec, RollingUpdateSpec
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    OffsetPagination,
    QueryCondition,
    QueryOrder,
)
from ai.backend.manager.repositories.deployment.options import (
    DeploymentConditions,
    DeploymentOrders,
    RevisionConditions,
    RevisionOrders,
    RouteConditions,
    RouteOrders,
)

__all__ = (
    "CreateDeploymentAdapter",
    "CreateRevisionAdapter",
    "DeploymentAdapter",
    "RevisionAdapter",
    "RouteAdapter",
)


class DeploymentAdapter(BaseFilterAdapter):
    """Adapter for converting deployment requests to repository queries."""

    def convert_to_dto(self, data: ModelDeploymentData) -> DeploymentDTO:
        """Convert ModelDeploymentData to DTO."""
        current_revision = None
        if data.revision:
            revision_adapter = RevisionAdapter()
            current_revision = revision_adapter.convert_to_dto(data.revision)

        return DeploymentDTO(
            id=data.id,
            name=data.metadata.name,
            status=data.metadata.status,
            tags=data.metadata.tags,
            project_id=data.metadata.project_id,
            domain_name=data.metadata.domain_name,
            created_at=data.metadata.created_at,
            updated_at=data.metadata.updated_at,
            created_user_id=data.created_user_id,
            network_config=NetworkConfigDTO(
                open_to_public=data.network_access.open_to_public,
                url=data.network_access.url,
                preferred_domain_name=data.network_access.preferred_domain_name,
            ),
            replica_state=ReplicaStateDTO(
                desired_replica_count=data.replica_state.desired_replica_count,
                replica_ids=data.replica_state.replica_ids,
            ),
            default_deployment_strategy=data.default_deployment_strategy,
            current_revision=current_revision,
        )

    def build_querier(self, request: SearchDeploymentsRequest) -> BatchQuerier:
        """
        Build a BatchQuerier for deployments from search request.

        Args:
            request: Search request containing filter, order, and pagination

        Returns:
            BatchQuerier object with converted conditions, orders, and pagination
        """
        conditions = self._convert_filter(request.filter) if request.filter else []
        orders = [self._convert_order(request.order)] if request.order else []
        pagination = self._build_pagination(request.limit, request.offset)

        return BatchQuerier(conditions=conditions, orders=orders, pagination=pagination)

    def _convert_filter(self, filter: DeploymentFilter) -> list[QueryCondition]:
        """Convert deployment filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        # Name filter
        if filter.name is not None:
            condition = self.convert_string_filter(
                filter.name,
                contains_factory=DeploymentConditions.by_name_contains,
                equals_factory=DeploymentConditions.by_name_equals,
                starts_with_factory=DeploymentConditions.by_name_starts_with,
                ends_with_factory=DeploymentConditions.by_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        # Domain name filter
        if filter.domain_name is not None:
            condition = self.convert_string_filter(
                filter.domain_name,
                contains_factory=DeploymentConditions.by_domain_name_contains,
                equals_factory=DeploymentConditions.by_domain_name_equals,
                starts_with_factory=DeploymentConditions.by_domain_name_starts_with,
                ends_with_factory=DeploymentConditions.by_domain_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        # Project ID filter
        if filter.project_id is not None:
            conditions.append(DeploymentConditions.by_project_id(filter.project_id))

        return conditions

    def _convert_order(self, order: DeploymentOrder) -> QueryOrder:
        """Convert deployment order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        if order.field == DeploymentOrderField.NAME:
            return DeploymentOrders.name(ascending=ascending)
        if order.field == DeploymentOrderField.CREATED_AT:
            return DeploymentOrders.created_at(ascending=ascending)
        raise ValueError(f"Unknown order field: {order.field}")

    def _build_pagination(self, limit: int, offset: int) -> OffsetPagination:
        """Build pagination from limit and offset."""
        return OffsetPagination(limit=limit, offset=offset)


class RevisionAdapter(BaseFilterAdapter):
    """Adapter for converting revision requests to repository queries."""

    def convert_to_dto(self, data: ModelRevisionData) -> RevisionDTO:
        """Convert ModelRevisionData to DTO."""
        return RevisionDTO(
            id=data.id,
            name=data.name,
            cluster_config=ClusterConfigDTO(
                mode=data.cluster_config.mode,
                size=data.cluster_config.size,
            ),
            resource_config=ResourceConfigDTO(
                resource_group_name=data.resource_config.resource_group_name,
                resource_slot=dict(data.resource_config.resource_slot),
            ),
            model_runtime_config=ModelRuntimeConfigDTO(
                runtime_variant=data.model_runtime_config.runtime_variant,
            ),
            model_mount_config=ModelMountConfigDTO(
                vfolder_id=data.model_mount_config.vfolder_id or uuid4(),
                mount_destination=data.model_mount_config.mount_destination or "",
                definition_path=data.model_mount_config.definition_path,
            ),
            created_at=data.created_at,
            image_id=data.image_id,
        )

    def build_querier(self, request: SearchRevisionsRequest) -> BatchQuerier:
        """
        Build a BatchQuerier for revisions from search request.

        Args:
            request: Search request containing filter, order, and pagination

        Returns:
            BatchQuerier object with converted conditions, orders, and pagination
        """
        conditions = self._convert_filter(request.filter) if request.filter else []
        orders = [self._convert_order(request.order)] if request.order else []
        pagination = self._build_pagination(request.limit, request.offset)

        return BatchQuerier(conditions=conditions, orders=orders, pagination=pagination)

    def _convert_filter(self, filter: RevisionFilter) -> list[QueryCondition]:
        """Convert revision filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        # Name filter
        if filter.name is not None:
            condition = self.convert_string_filter(
                filter.name,
                contains_factory=RevisionConditions.by_name_contains,
                equals_factory=RevisionConditions.by_name_equals,
                starts_with_factory=RevisionConditions.by_name_starts_with,
                ends_with_factory=RevisionConditions.by_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        # Deployment ID filter
        if filter.deployment_id is not None:
            conditions.append(RevisionConditions.by_deployment_id(filter.deployment_id))

        return conditions

    def _convert_order(self, order: RevisionOrder) -> QueryOrder:
        """Convert revision order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        if order.field == RevisionOrderField.NAME:
            return RevisionOrders.name(ascending=ascending)
        if order.field == RevisionOrderField.CREATED_AT:
            return RevisionOrders.created_at(ascending=ascending)
        raise ValueError(f"Unknown order field: {order.field}")

    def _build_pagination(self, limit: int, offset: int) -> OffsetPagination:
        """Build pagination from limit and offset."""
        return OffsetPagination(limit=limit, offset=offset)


class RouteAdapter(BaseFilterAdapter):
    """Adapter for converting route requests to repository queries."""

    def convert_to_dto(self, data: RouteInfo) -> RouteDTO:
        """Convert RouteInfo to DTO."""
        return RouteDTO(
            id=data.route_id,
            endpoint_id=data.endpoint_id,
            session_id=str(data.session_id) if data.session_id else None,
            status=CommonRouteStatus(data.status.value),
            traffic_ratio=data.traffic_ratio,
            created_at=data.created_at or datetime.now(tz=UTC),
            revision_id=data.revision_id,
            traffic_status=CommonRouteTrafficStatus(data.traffic_status.value),
            error_data=data.error_data,
        )

    def build_querier(self, request: SearchRoutesRequest) -> BatchQuerier:
        """
        Build a BatchQuerier for routes from search request.

        Args:
            request: Search request containing filter, order, and pagination

        Returns:
            BatchQuerier object with converted conditions, orders, and pagination
        """
        conditions = self._convert_filter(request.filter) if request.filter else []
        orders = [self._convert_order(request.order)] if request.order else []

        # Add cursor conditions if provided
        if request.cursor:
            if request.cursor_direction == "forward":
                conditions.append(RouteConditions.by_cursor_forward(request.cursor))
            elif request.cursor_direction == "backward":
                conditions.append(RouteConditions.by_cursor_backward(request.cursor))

        pagination = self._build_pagination(request.limit, request.offset)

        return BatchQuerier(conditions=conditions, orders=orders, pagination=pagination)

    def _convert_filter(self, filter: RouteFilter) -> list[QueryCondition]:
        """Convert route filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        # Deployment ID filter
        if filter.deployment_id is not None:
            conditions.append(RouteConditions.by_endpoint_id(filter.deployment_id))

        # Status filter - convert common types to manager types
        if filter.statuses is not None:
            manager_statuses = [ManagerRouteStatus(s.value) for s in filter.statuses]
            conditions.append(RouteConditions.by_statuses(manager_statuses))

        # Traffic status filter - convert common types to manager types
        if filter.traffic_statuses is not None:
            manager_traffic_statuses = [
                ManagerRouteTrafficStatus(s.value) for s in filter.traffic_statuses
            ]
            conditions.append(RouteConditions.by_traffic_statuses(manager_traffic_statuses))

        return conditions

    def _convert_order(self, order: RouteOrder) -> QueryOrder:
        """Convert route order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        if order.field == RouteOrderField.CREATED_AT:
            return RouteOrders.created_at(ascending=ascending)
        if order.field == RouteOrderField.STATUS:
            return RouteOrders.status(ascending=ascending)
        if order.field == RouteOrderField.TRAFFIC_RATIO:
            return RouteOrders.traffic_ratio(ascending=ascending)
        raise ValueError(f"Unknown order field: {order.field}")

    def _build_pagination(self, limit: int, offset: int) -> OffsetPagination:
        """Build pagination from limit and offset."""
        return OffsetPagination(limit=limit, offset=offset)


class CreateDeploymentAdapter:
    """Adapter for converting create deployment request to creators."""

    def build_creator(
        self,
        request: CreateDeploymentRequest,
        user_uuid: UUID,
    ) -> NewDeploymentCreator:
        """
        Convert CreateDeploymentRequest to NewDeploymentCreator.

        Args:
            request: Create deployment request DTO
            user_uuid: UUID of the user creating the deployment

        Returns:
            NewDeploymentCreator for service layer
        """
        # Generate name if not provided
        name = request.metadata.name or f"deployment-{uuid4().hex[:8]}"
        tag = ",".join(request.metadata.tags) if request.metadata.tags else None

        # Build metadata
        metadata = DeploymentMetadata(
            name=name,
            domain=request.metadata.domain_name,
            project=request.metadata.project_id,
            resource_group=request.initial_revision.resource_config.resource_group,
            created_user=user_uuid,
            session_owner=user_uuid,
            created_at=None,
            revision_history_limit=10,
            tag=tag,
        )

        # Build replica spec
        replica_spec = ReplicaSpec(replica_count=request.desired_replica_count)

        # Build network spec
        network = DeploymentNetworkSpec(
            open_to_public=request.network_access.open_to_public,
            preferred_domain_name=request.network_access.preferred_domain_name,
        )

        # Build model revision creator
        model_revision = self._build_revision_creator(request.initial_revision)

        # Build policy config
        policy = self._build_policy_config(request.default_deployment_strategy)

        return NewDeploymentCreator(
            metadata=metadata,
            replica_spec=replica_spec,
            network=network,
            model_revision=model_revision,
            policy=policy,
        )

    def _build_revision_creator(
        self,
        revision_input: Any,  # RevisionInput or CreateRevisionRequest
    ) -> ModelRevisionCreator:
        """Build ModelRevisionCreator from revision input."""
        # Build resource spec
        resource_spec = ResourceSpec(
            cluster_mode=ClusterMode(revision_input.cluster_config.mode),
            cluster_size=revision_input.cluster_config.size,
            resource_slots=dict(revision_input.resource_config.resource_slots),
            resource_opts=(
                dict(revision_input.resource_config.resource_opts)
                if revision_input.resource_config.resource_opts
                else None
            ),
        )

        # Build extra mounts
        extra_mounts: list[MountInfo] = []
        if revision_input.extra_mounts:
            extra_mounts = [
                MountInfo(
                    vfolder_id=mount.vfolder_id,
                    kernel_path=PurePosixPath(mount.mount_destination or ""),
                )
                for mount in revision_input.extra_mounts
            ]

        # Build vfolder mounts creator
        mounts = VFolderMountsCreator(
            model_vfolder_id=revision_input.model_mount_config.vfolder_id,
            model_definition_path=revision_input.model_mount_config.definition_path,
            model_mount_destination=revision_input.model_mount_config.mount_destination,
            extra_mounts=extra_mounts,
        )

        # Build execution spec
        execution = ExecutionSpec(
            runtime_variant=RuntimeVariant(revision_input.model_runtime_config.runtime_variant),
            inference_runtime_config=(
                dict(revision_input.model_runtime_config.inference_runtime_config)
                if revision_input.model_runtime_config.inference_runtime_config
                else None
            ),
            environ=(
                dict(revision_input.model_runtime_config.environ)
                if revision_input.model_runtime_config.environ
                else None
            ),
        )

        return ModelRevisionCreator(
            image_id=revision_input.image.id,
            resource_spec=resource_spec,
            mounts=mounts,
            execution=execution,
        )

    def _build_policy_config(
        self,
        strategy_input: Any,  # DeploymentStrategyInput
    ) -> DeploymentPolicyConfig:
        """Build DeploymentPolicyConfig from strategy input."""
        strategy = DeploymentStrategy(strategy_input.type)

        strategy_spec: RollingUpdateSpec | BlueGreenSpec
        match strategy:
            case DeploymentStrategy.ROLLING:
                if strategy_input.rolling_update is None:
                    strategy_spec = RollingUpdateSpec(max_surge=1, max_unavailable=0)
                else:
                    strategy_spec = RollingUpdateSpec(
                        max_surge=strategy_input.rolling_update.max_surge,
                        max_unavailable=strategy_input.rolling_update.max_unavailable,
                    )
            case DeploymentStrategy.BLUE_GREEN:
                if strategy_input.blue_green is None:
                    strategy_spec = BlueGreenSpec(auto_promote=False, promote_delay_seconds=0)
                else:
                    strategy_spec = BlueGreenSpec(
                        auto_promote=strategy_input.blue_green.auto_promote,
                        promote_delay_seconds=strategy_input.blue_green.promote_delay_seconds,
                    )

        return DeploymentPolicyConfig(
            strategy=strategy,
            strategy_spec=strategy_spec,
            rollback_on_failure=strategy_input.rollback_on_failure,
        )


class CreateRevisionAdapter:
    """Adapter for converting create revision request to creators."""

    def build_creator(self, request: CreateRevisionRequest) -> ModelRevisionCreator:
        """
        Convert CreateRevisionRequest to ModelRevisionCreator.

        Args:
            request: Create revision request DTO

        Returns:
            ModelRevisionCreator for service layer
        """
        deployment_adapter = CreateDeploymentAdapter()
        return deployment_adapter._build_revision_creator(request)



"""
Repository layer exceptions.
"""

from __future__ import annotations

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class RepositoryError(BackendAIError):
    """Base class for repository layer errors."""

    error_type = "https://api.backend.ai/probs/repository-error"
    error_title = "Repository operation failed."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UpsertEmptyResultError(RepositoryError):
    """Raised when upsert operation returns no rows."""

    error_type = "https://api.backend.ai/probs/upsert-empty-result"
    error_title = "Upsert operation did not return any row."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UnsupportedCompositePrimaryKeyError(RepositoryError):
    """Raised when an operation requires a single-column primary key but the table has a composite key."""

    error_type = "https://api.backend.ai/probs/unsupported-composite-pk"
    error_title = "Unsupported composite primary key."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )



"""Keypair export report definition."""

from __future__ import annotations

import json
from decimal import Decimal
from typing import Any

from ai.backend.manager.models.keypair import KeyPairRow
from ai.backend.manager.models.resource_policy import KeyPairResourcePolicyRow
from ai.backend.manager.models.scaling_group import ScalingGroupForKeypairsRow, ScalingGroupRow
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base.export import (
    ExportFieldDef,
    ExportFieldType,
    JoinDef,
    ReportDef,
)

# =============================================================================
# Helper Functions
# =============================================================================


def _serialize_json(value: Any) -> str:
    """Serialize value to JSON string, converting sets to lists and Decimals to strings."""
    if not value:
        return ""

    def convert(obj: Any) -> Any:
        if isinstance(obj, Decimal):
            return str(obj)
        if isinstance(obj, set):
            return sorted(obj)
        if isinstance(obj, dict):
            return {k: convert(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [convert(i) for i in obj]
        return obj

    return json.dumps(convert(dict(value)))


# =============================================================================
# JOIN Definitions
# =============================================================================

# User JOIN (N:1, no duplication)
USER_JOIN = JoinDef(
    table=UserRow.__table__,
    condition=KeyPairRow.user == UserRow.uuid,
)

# Resource Policy JOIN (N:1, no duplication)
RESOURCE_POLICY_JOIN = JoinDef(
    table=KeyPairResourcePolicyRow.__table__,
    condition=KeyPairRow.resource_policy == KeyPairResourcePolicyRow.name,
)

# Resource Group JOINs (1:N, causes duplication)
SGROUP_FOR_KEYPAIR_JOIN = JoinDef(
    table=ScalingGroupForKeypairsRow.__table__,
    condition=KeyPairRow.access_key == ScalingGroupForKeypairsRow.access_key,
)
RESOURCE_GROUP_JOIN = JoinDef(
    table=ScalingGroupRow.__table__,
    condition=ScalingGroupForKeypairsRow.scaling_group == ScalingGroupRow.name,
)
RESOURCE_GROUP_JOINS = (SGROUP_FOR_KEYPAIR_JOIN, RESOURCE_GROUP_JOIN)

# Session JOIN (1:N, causes duplication)
SESSION_JOIN = JoinDef(
    table=SessionRow.__table__,
    condition=KeyPairRow.access_key == SessionRow.access_key,
)

# Field definitions for keypair export
KEYPAIR_FIELDS: list[ExportFieldDef] = [
    ExportFieldDef(
        key="access_key",
        name="Access Key",
        description="Access key identifier",
        field_type=ExportFieldType.STRING,
        column=KeyPairRow.access_key,
    ),
    ExportFieldDef(
        key="is_active",
        name="Active",
        description="Keypair active status",
        field_type=ExportFieldType.BOOLEAN,
        column=KeyPairRow.is_active,
    ),
    ExportFieldDef(
        key="is_admin",
        name="Admin",
        description="Admin keypair status",
        field_type=ExportFieldType.BOOLEAN,
        column=KeyPairRow.is_admin,
    ),
    ExportFieldDef(
        key="created_at",
        name="Created At",
        description="Keypair creation time",
        field_type=ExportFieldType.DATETIME,
        column=KeyPairRow.created_at,
        formatter=lambda v: v.isoformat() if v else "",
    ),
    ExportFieldDef(
        key="modified_at",
        name="Modified At",
        description="Last modification time",
        field_type=ExportFieldType.DATETIME,
        column=KeyPairRow.modified_at,
        formatter=lambda v: v.isoformat() if v else "",
    ),
    ExportFieldDef(
        key="last_used",
        name="Last Used",
        description="Last used timestamp",
        field_type=ExportFieldType.DATETIME,
        column=KeyPairRow.last_used,
        formatter=lambda v: v.isoformat() if v else "",
    ),
    # =========================================================================
    # User Fields (N:1, no duplication)
    # =========================================================================
    ExportFieldDef(
        key="user_id",
        name="User Email",
        description="User email (ID)",
        field_type=ExportFieldType.STRING,
        column=UserRow.email,
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_uuid",
        name="User UUID",
        description="User UUID",
        field_type=ExportFieldType.UUID,
        column=UserRow.uuid,
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_username",
        name="User Username",
        description="Username",
        field_type=ExportFieldType.STRING,
        column=UserRow.username,
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_full_name",
        name="User Full Name",
        description="User full name",
        field_type=ExportFieldType.STRING,
        column=UserRow.full_name,
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_role",
        name="User Role",
        description="User role",
        field_type=ExportFieldType.STRING,
        column=UserRow.role,
        formatter=lambda v: str(v) if v else "",
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_status",
        name="User Status",
        description="User account status",
        field_type=ExportFieldType.STRING,
        column=UserRow.status,
        formatter=lambda v: str(v) if v else "",
        joins=frozenset({USER_JOIN}),
    ),
    ExportFieldDef(
        key="user_domain_name",
        name="User Domain",
        description="User domain name",
        field_type=ExportFieldType.STRING,
        column=UserRow.domain_name,
        joins=frozenset({USER_JOIN}),
    ),
    # =========================================================================
    # Keypair Resource Policy Fields (N:1, no duplication)
    # =========================================================================
    ExportFieldDef(
        key="resource_policy_name",
        name="Resource Policy Name",
        description="Keypair resource policy name",
        field_type=ExportFieldType.STRING,
        column=KeyPairResourcePolicyRow.name,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_created_at",
        name="Policy Created At",
        description="Resource policy creation time",
        field_type=ExportFieldType.DATETIME,
        column=KeyPairResourcePolicyRow.created_at,
        formatter=lambda v: v.isoformat() if v else "",
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_max_concurrent_sessions",
        name="Policy Max Concurrent Sessions",
        description="Maximum concurrent sessions allowed",
        field_type=ExportFieldType.INTEGER,
        column=KeyPairResourcePolicyRow.max_concurrent_sessions,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_max_containers_per_session",
        name="Policy Max Containers Per Session",
        description="Maximum containers per session",
        field_type=ExportFieldType.INTEGER,
        column=KeyPairResourcePolicyRow.max_containers_per_session,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_idle_timeout",
        name="Policy Idle Timeout",
        description="Idle timeout in seconds",
        field_type=ExportFieldType.INTEGER,
        column=KeyPairResourcePolicyRow.idle_timeout,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_max_session_lifetime",
        name="Policy Max Session Lifetime",
        description="Maximum session lifetime in seconds",
        field_type=ExportFieldType.INTEGER,
        column=KeyPairResourcePolicyRow.max_session_lifetime,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_max_pending_session_count",
        name="Policy Max Pending Sessions",
        description="Maximum number of pending sessions",
        field_type=ExportFieldType.INTEGER,
        column=KeyPairResourcePolicyRow.max_pending_session_count,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_max_pending_session_resource_slots",
        name="Policy Max Pending Session Resources",
        description="Maximum resource slots for pending sessions",
        field_type=ExportFieldType.JSON,
        column=KeyPairResourcePolicyRow.max_pending_session_resource_slots,
        formatter=_serialize_json,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    ExportFieldDef(
        key="resource_policy_max_concurrent_sftp_sessions",
        name="Policy Max Concurrent SFTP Sessions",
        description="Maximum concurrent SFTP sessions",
        field_type=ExportFieldType.INTEGER,
        column=KeyPairResourcePolicyRow.max_concurrent_sftp_sessions,
        joins=frozenset({RESOURCE_POLICY_JOIN}),
    ),
    # =========================================================================
    # Resource Group Fields (1:N, causes duplication)
    # =========================================================================
    ExportFieldDef(
        key="resource_group_name",
        name="Resource Group Name",
        description="Resource group name",
        field_type=ExportFieldType.STRING,
        column=ScalingGroupRow.name,
        joins=RESOURCE_GROUP_JOINS,
    ),
    ExportFieldDef(
        key="resource_group_description",
        name="Resource Group Description",
        description="Resource group description",
        field_type=ExportFieldType.STRING,
        column=ScalingGroupRow.description,
        joins=RESOURCE_GROUP_JOINS,
    ),
    ExportFieldDef(
        key="resource_group_is_active",
        name="Resource Group Active",
        description="Resource group active status",
        field_type=ExportFieldType.BOOLEAN,
        column=ScalingGroupRow.is_active,
        joins=RESOURCE_GROUP_JOINS,
    ),
    ExportFieldDef(
        key="resource_group_scheduler",
        name="Resource Group Scheduler",
        description="Resource group scheduler type",
        field_type=ExportFieldType.STRING,
        column=ScalingGroupRow.scheduler,
        joins=RESOURCE_GROUP_JOINS,
    ),
    ExportFieldDef(
        key="resource_group_wsproxy_addr",
        name="Resource Group WSProxy Address",
        description="WebSocket proxy address for resource group",
        field_type=ExportFieldType.STRING,
        column=ScalingGroupRow.wsproxy_addr,
        joins=RESOURCE_GROUP_JOINS,
    ),
    ExportFieldDef(
        key="resource_group_fair_share_spec",
        name="Resource Group Fair Share Spec",
        description="Fair share specification for resource group",
        field_type=ExportFieldType.JSON,
        column=ScalingGroupRow.fair_share_spec,
        formatter=_serialize_json,
        joins=RESOURCE_GROUP_JOINS,
    ),
    # =========================================================================
    # Session Fields (1:N, causes duplication)
    # =========================================================================
    ExportFieldDef(
        key="session_id",
        name="Session ID",
        description="Session UUID",
        field_type=ExportFieldType.UUID,
        column=SessionRow.id,
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_name",
        name="Session Name",
        description="Session name",
        field_type=ExportFieldType.STRING,
        column=SessionRow.name,
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_status",
        name="Session Status",
        description="Session status",
        field_type=ExportFieldType.STRING,
        column=SessionRow.status,
        formatter=lambda v: str(v) if v else "",
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_type",
        name="Session Type",
        description="Session type (interactive, batch, etc.)",
        field_type=ExportFieldType.STRING,
        column=SessionRow.session_type,
        formatter=lambda v: str(v) if v else "",
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_domain_name",
        name="Session Domain",
        description="Session domain name",
        field_type=ExportFieldType.STRING,
        column=SessionRow.domain_name,
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_group_id",
        name="Session Group ID",
        description="Session project/group UUID",
        field_type=ExportFieldType.UUID,
        column=SessionRow.group_id,
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_created_at",
        name="Session Created At",
        description="Session creation time",
        field_type=ExportFieldType.DATETIME,
        column=SessionRow.created_at,
        formatter=lambda v: v.isoformat() if v else "",
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_terminated_at",
        name="Session Terminated At",
        description="Session termination time",
        field_type=ExportFieldType.DATETIME,
        column=SessionRow.terminated_at,
        formatter=lambda v: v.isoformat() if v else "",
        joins=frozenset({SESSION_JOIN}),
    ),
    ExportFieldDef(
        key="session_resource_requested",
        name="Session Resource Requested",
        description="Requested resource slots for session",
        field_type=ExportFieldType.JSON,
        column=SessionRow.requested_slots,
        formatter=_serialize_json,
        joins=frozenset({SESSION_JOIN}),
    ),
]


# Report definition
KEYPAIR_REPORT = ReportDef(
    report_key="keypairs",
    name="Keypairs",
    description="Keypair (access key) export report",
    select_from=KeyPairRow.__table__,
    fields=KEYPAIR_FIELDS,
)



import asyncio
import logging
from collections.abc import Callable

import sqlalchemy as sa

from ai.backend.common.events.event_types.agent.anycast import (
    AgentErrorEvent,
    AgentHeartbeatEvent,
    AgentImagesRemoveEvent,
    AgentInstalledImagesRemoveEvent,
    AgentStartedEvent,
    AgentTerminatedEvent,
    DoAgentResourceCheckEvent,
)
from ai.backend.common.exception import ProcessorNotReadyError
from ai.backend.common.plugin.event import EventDispatcherPluginContext
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.resource import InstanceNotFound
from ai.backend.manager.models.agent import AgentStatus, agents
from ai.backend.manager.models.resource_slot import AgentResourceRow
from ai.backend.manager.models.utils import (
    ExtendedAsyncSAEngine,
)
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.services.agent.actions.handle_heartbeat import HandleHeartbeatAction
from ai.backend.manager.services.agent.actions.mark_agent_exit import MarkAgentExitAction
from ai.backend.manager.services.agent.actions.mark_agent_running import MarkAgentRunningAction
from ai.backend.manager.services.agent.actions.remove_agent_from_images import (
    RemoveAgentFromImagesAction,
)
from ai.backend.manager.services.agent.actions.remove_agent_from_images_by_canonicals import (
    RemoveAgentFromImagesByCanonicalsAction,
)
from ai.backend.manager.services.processors import Processors

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AgentEventHandler:
    _registry: AgentRegistry
    _db: ExtendedAsyncSAEngine
    _event_dispatcher_plugin_ctx: EventDispatcherPluginContext
    _processor_factory: Callable[[], Processors]
    _processors: Processors | None

    def __init__(
        self,
        registry: AgentRegistry,
        db: ExtendedAsyncSAEngine,
        event_dispatcher_plugin_ctx: EventDispatcherPluginContext,
        processors_factory: Callable[[], Processors],
    ) -> None:
        self._registry = registry
        self._db = db
        self._event_dispatcher_plugin_ctx = event_dispatcher_plugin_ctx
        self._processors = None
        self._processor_factory = processors_factory

    # Lazy initialization of processors as AgentEventHandler is created earlier than Processors
    async def get_processors(self) -> Processors:
        if self._processors is None:
            for _ in range(5):
                try:
                    self._processors = self._processor_factory()
                    return self._processors
                except Exception:
                    await asyncio.sleep(0.1)
        if self._processors is None:
            log.error("Agent processors not ready after multiple attempts.")
            raise ProcessorNotReadyError("Agent processors not ready. Try again after a while.")
        return self._processors

    async def handle_agent_started(
        self,
        _context: None,
        source: AgentId,
        event: AgentStartedEvent,
    ) -> None:
        log.info("instance_lifecycle: ag:{0} joined (via event, {1})", source, event.reason)
        processors = await self.get_processors()
        await processors.agent.mark_agent_running.wait_for_complete(
            MarkAgentRunningAction(
                agent_id=source,
                agent_status=AgentStatus.ALIVE,
            )
        )

    async def handle_agent_terminated(
        self,
        _context: None,
        source: AgentId,
        event: AgentTerminatedEvent,
    ) -> None:
        processors = await self.get_processors()
        if event.reason == "agent-lost":
            await processors.agent.mark_agent_exit.wait_for_complete(
                MarkAgentExitAction(
                    agent_id=source,
                    agent_status=AgentStatus.LOST,
                )
            )
        elif event.reason == "agent-restart":
            log.info("agent@{0} restarting for maintenance.", source)
            await processors.agent.mark_agent_running.wait_for_complete(
                MarkAgentRunningAction(
                    agent_id=source,
                    agent_status=AgentStatus.RESTARTING,
                )
            )
        else:
            # On normal instance termination, kernel_terminated events were already
            # triggered by the agent.
            await processors.agent.mark_agent_exit.wait_for_complete(
                MarkAgentExitAction(
                    agent_id=source,
                    agent_status=AgentStatus.TERMINATED,
                )
            )

    async def handle_agent_heartbeat(
        self,
        _context: None,
        source: AgentId,
        event: AgentHeartbeatEvent,
    ) -> None:
        processor = await self.get_processors()
        await processor.agent.handle_heartbeat.wait_for_complete(
            action=HandleHeartbeatAction(agent_id=source, agent_info=event.agent_info)
        )

    # For compatibility with redis key made with image canonical strings
    # Use remove_agent_from_images_by_id instead of this if possible
    async def handle_agent_images_remove(
        self,
        _context: None,
        source: AgentId,
        event: AgentImagesRemoveEvent,
    ) -> None:
        processor = await self.get_processors()
        await processor.agent.remove_agent_from_images_by_canonicals.wait_for_complete(
            action=RemoveAgentFromImagesByCanonicalsAction(
                agent_id=source, image_canonicals=event.image_canonicals
            )
        )

    async def handle_agent_installed_images_remove(
        self,
        _context: None,
        source: AgentId,
        event: AgentInstalledImagesRemoveEvent,
    ) -> None:
        processor = await self.get_processors()
        await processor.agent.remove_agent_from_images.wait_for_complete(
            action=RemoveAgentFromImagesAction(
                agent_id=source, scanned_images=dict(event.scanned_images)
            )
        )

    async def handle_check_agent_resource(
        self, _context: None, source: AgentId, _event: DoAgentResourceCheckEvent
    ) -> None:
        async with self._db.begin_readonly() as conn:
            # Check agent existence
            agent_query = sa.select(sa.literal(1)).select_from(agents).where(agents.c.id == source)
            agent_result = await conn.execute(agent_query)
            if agent_result.first() is None:
                raise InstanceNotFound(source)
            # Read used slots from normalized agent_resources table
            ar = AgentResourceRow.__table__
            query = sa.select(ar.c.slot_name, ar.c.used).where(ar.c.agent_id == source)
            result = await conn.execute(query)
            used_slots = {row.slot_name: row.used for row in result}
            log.info("agent@{0} used slots: {1}", source, used_slots)

    async def handle_agent_error(
        self,
        context: None,
        source: AgentId,
        event: AgentErrorEvent,
    ) -> None:
        await self._event_dispatcher_plugin_ctx.handle_event(context, source, event)



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec

from .actions import (
    AssignRoleAction,
    AssignRoleActionResult,
    CreateRoleAction,
    CreateRoleActionResult,
    DeleteRoleAction,
    DeleteRoleActionResult,
    GetRoleDetailAction,
    GetRoleDetailActionResult,
    PurgeRoleAction,
    RevokeRoleAction,
    RevokeRoleActionResult,
    SearchRolesAction,
    SearchRolesActionResult,
    SearchUsersAssignedToRoleAction,
    SearchUsersAssignedToRoleActionResult,
    UpdateRoleAction,
    UpdateRoleActionResult,
    UpdateRolePermissionsAction,
    UpdateRolePermissionsActionResult,
)
from .actions.get_entity_types import (
    GetEntityTypesAction,
    GetEntityTypesActionResult,
)
from .actions.get_scope_types import (
    GetScopeTypesAction,
    GetScopeTypesActionResult,
)
from .actions.permission import (
    CreatePermissionAction,
    CreatePermissionActionResult,
    DeletePermissionAction,
    DeletePermissionActionResult,
)
from .actions.search_entities import (
    SearchEntitiesAction,
    SearchEntitiesActionResult,
)
from .actions.search_permissions import (
    SearchPermissionsAction,
    SearchPermissionsActionResult,
)
from .actions.search_scopes import (
    SearchScopesAction,
    SearchScopesActionResult,
)
from .service import PermissionControllerService


class PermissionControllerProcessors(AbstractProcessorPackage):
    """Processor package for RBAC permission controller operations."""

    create_role: ActionProcessor[CreateRoleAction, CreateRoleActionResult]
    update_role: ActionProcessor[UpdateRoleAction, UpdateRoleActionResult]
    delete_role: ActionProcessor[DeleteRoleAction, DeleteRoleActionResult]
    assign_role: ActionProcessor[AssignRoleAction, AssignRoleActionResult]
    revoke_role: ActionProcessor[RevokeRoleAction, RevokeRoleActionResult]
    get_role_detail: ActionProcessor[GetRoleDetailAction, GetRoleDetailActionResult]
    search_roles: ActionProcessor[SearchRolesAction, SearchRolesActionResult]
    search_users_assigned_to_role: ActionProcessor[
        SearchUsersAssignedToRoleAction, SearchUsersAssignedToRoleActionResult
    ]
    update_role_permissions: ActionProcessor[
        UpdateRolePermissionsAction, UpdateRolePermissionsActionResult
    ]
    search_scopes: ActionProcessor[SearchScopesAction, SearchScopesActionResult]
    get_scope_types: ActionProcessor[GetScopeTypesAction, GetScopeTypesActionResult]
    get_entity_types: ActionProcessor[GetEntityTypesAction, GetEntityTypesActionResult]
    search_entities: ActionProcessor[SearchEntitiesAction, SearchEntitiesActionResult]
    search_permissions: ActionProcessor[SearchPermissionsAction, SearchPermissionsActionResult]
    create_permission: ActionProcessor[CreatePermissionAction, CreatePermissionActionResult]
    delete_permission: ActionProcessor[DeletePermissionAction, DeletePermissionActionResult]

    def __init__(
        self, service: PermissionControllerService, action_monitors: list[ActionMonitor]
    ) -> None:
        self.create_role = ActionProcessor(service.create_role, action_monitors)
        self.update_role = ActionProcessor(service.update_role, action_monitors)
        self.delete_role = ActionProcessor(service.delete_role, action_monitors)
        self.purge_role = ActionProcessor(service.purge_role, action_monitors)
        self.assign_role = ActionProcessor(service.assign_role, action_monitors)
        self.revoke_role = ActionProcessor(service.revoke_role, action_monitors)
        self.get_role_detail = ActionProcessor(service.get_role_detail, action_monitors)
        self.search_roles = ActionProcessor(service.search_roles, action_monitors)
        self.search_users_assigned_to_role = ActionProcessor(
            service.search_users_assigned_to_role, action_monitors
        )
        self.update_role_permissions = ActionProcessor(
            service.update_role_permissions, action_monitors
        )
        self.search_scopes = ActionProcessor(service.search_scopes, action_monitors)
        self.get_scope_types = ActionProcessor(service.get_scope_types, action_monitors)
        self.get_entity_types = ActionProcessor(service.get_entity_types, action_monitors)
        self.search_entities = ActionProcessor(service.search_entities, action_monitors)
        self.search_permissions = ActionProcessor(service.search_permissions, action_monitors)
        self.create_permission = ActionProcessor(service.create_permission, action_monitors)
        self.delete_permission = ActionProcessor(service.delete_permission, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateRoleAction.spec(),
            UpdateRoleAction.spec(),
            DeleteRoleAction.spec(),
            PurgeRoleAction.spec(),
            AssignRoleAction.spec(),
            RevokeRoleAction.spec(),
            GetRoleDetailAction.spec(),
            SearchRolesAction.spec(),
            SearchUsersAssignedToRoleAction.spec(),
            UpdateRolePermissionsAction.spec(),
            SearchScopesAction.spec(),
            GetScopeTypesAction.spec(),
            GetEntityTypesAction.spec(),
            SearchEntitiesAction.spec(),
            SearchPermissionsAction.spec(),
            CreatePermissionAction.spec(),
            DeletePermissionAction.spec(),
        ]



from __future__ import annotations

import uuid
from dataclasses import dataclass

from ai.backend.manager.data.common.types import SearchResult
from ai.backend.manager.types import OptionalState

from .id import ObjectId
from .status import PermissionStatus
from .types import EntityType, OperationType


@dataclass
class ObjectPermissionCreateInputBeforeRoleCreation:
    """Input data for creating ObjectPermission before role is created.

    This is a plain data container. The conversion to CreatorSpec happens
    in the repository layer (db_source.py) to maintain proper dependency direction.
    """

    entity_type: EntityType
    entity_id: str
    operation: OperationType
    status: PermissionStatus = PermissionStatus.ACTIVE


@dataclass
class ObjectPermissionCreateInput:
    """Input data for creating ObjectPermission for an existing role.

    Used when adding object permissions to a role that already exists.
    """

    entity_type: EntityType
    entity_id: str
    operation: OperationType
    status: PermissionStatus = PermissionStatus.ACTIVE


@dataclass
class ObjectPermissionUpdater:
    id: uuid.UUID
    status: OptionalState[PermissionStatus]


@dataclass
class ObjectPermissionDeleteInput:
    id: uuid.UUID
    _status: PermissionStatus = PermissionStatus.DELETED


@dataclass
class ObjectPermissionData:
    id: uuid.UUID
    role_id: uuid.UUID
    object_id: ObjectId
    operation: OperationType


@dataclass(frozen=True)
class ObjectPermissionListResult(SearchResult[ObjectPermissionData]):
    """Result of object permission search with pagination info."""

    pass



from __future__ import annotations

import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import override

from ai.backend.manager.actions.types import OperationStatus
from ai.backend.manager.models.audit_log import AuditLogRow
from ai.backend.manager.repositories.base import CreatorSpec

__all__ = ("AuditLogCreatorSpec",)


@dataclass
class AuditLogCreatorSpec(CreatorSpec[AuditLogRow]):
    action_id: uuid.UUID
    entity_type: str
    operation: str
    created_at: datetime
    description: str
    status: OperationStatus
    entity_id: str | None
    request_id: str | None
    triggered_by: str | None
    duration: timedelta | None

    @override
    def build_row(self) -> AuditLogRow:
        return AuditLogRow(
            action_id=self.action_id,
            entity_type=self.entity_type,
            operation=self.operation,
            created_at=self.created_at,
            description=self.description,
            status=self.status,
            entity_id=self.entity_id,
            request_id=self.request_id,
            triggered_by=self.triggered_by,
            duration=self.duration,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.services.vfs_storage.actions.base import VFSStorageAction


@dataclass
class SearchVFSStoragesAction(VFSStorageAction):
    """Action to search VFS storages."""

    querier: BatchQuerier

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class SearchVFSStoragesActionResult(BaseActionResult):
    """Result of searching VFS storages."""

    storages: list[VFSStorageData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""Notification-related error definitions."""

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)

__all__ = (
    "InvalidNotificationChannelType",
    "InvalidNotificationSpec",
    "NotificationChannelConflict",
    "NotificationChannelNotFound",
    "NotificationProcessingFailure",
    "NotificationRuleConflict",
    "NotificationRuleNotFound",
    "NotificationTemplateRenderingFailure",
)


class NotificationChannelNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/notification-channel-not-found"
    error_title = "The notification channel does not exist."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class NotificationRuleNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/notification-rule-not-found"
    error_title = "The notification rule does not exist."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class NotificationChannelConflict(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/notification-channel-conflict"
    error_title = "The notification channel already exists."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.CONFLICT,
        )


class NotificationRuleConflict(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/notification-rule-conflict"
    error_title = "The notification rule already exists."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.CONFLICT,
        )


class NotificationProcessingFailure(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/notification-processing-failure"
    error_title = "Failed to process notification."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class NotificationTemplateRenderingFailure(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/notification-template-rendering-failure"
    error_title = "Failed to render notification template."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidNotificationChannelType(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-notification-channel-type"
    error_title = "Invalid notification channel type."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidNotificationSpec(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-notification-spec"
    error_title = "Invalid notification specification."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.NOTIFICATION,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )



"""
Adapter for Entity API requests.
Handles conversion of request DTOs to BatchQuerier objects.
"""

from __future__ import annotations

from ai.backend.common.data.permission.types import EntityType, ScopeType
from ai.backend.common.dto.manager.rbac.request import (
    SearchEntitiesRequest,
)
from ai.backend.common.dto.manager.rbac.response import EntityDTO
from ai.backend.manager.api.adapter import BaseFilterAdapter
from ai.backend.manager.data.permission.entity import EntityData
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    OffsetPagination,
)
from ai.backend.manager.repositories.permission_controller.options import (
    EntityScopeConditions,
)

__all__ = ("EntityAdapter",)


class EntityAdapter(BaseFilterAdapter):
    """Adapter for converting entity requests to BatchQuerier objects."""

    def build_querier(
        self,
        scope_type: ScopeType,
        scope_id: str,
        entity_type: EntityType,
        request: SearchEntitiesRequest,
    ) -> BatchQuerier:
        """Build a BatchQuerier for entity search.

        Args:
            scope_type: The scope type to search within
            scope_id: The scope ID to search within
            entity_type: The type of entity to search
            request: The search request containing pagination info

        Returns:
            BatchQuerier with scope conditions and pagination settings
        """
        conditions = [
            EntityScopeConditions.by_scope_type(scope_type),
            EntityScopeConditions.by_scope_id(scope_id),
            EntityScopeConditions.by_entity_type(entity_type),
        ]
        pagination = OffsetPagination(limit=request.limit, offset=request.offset)
        return BatchQuerier(conditions=conditions, orders=[], pagination=pagination)

    def convert_to_dto(self, data: EntityData) -> EntityDTO:
        """Convert EntityData to DTO.

        Args:
            data: Entity data from action result

        Returns:
            EntityDTO for API response
        """
        return EntityDTO(
            entity_type=data.entity_type,
            entity_id=data.entity_id,
        )



from __future__ import annotations

import logging
import uuid
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

import graphene
import sqlalchemy as sa
from graphene.types.datetime import DateTime as GQLDateTime
from graphql import Undefined
from sqlalchemy.engine.row import Row
from sqlalchemy.orm import selectinload

from ai.backend.common.types import DefaultForUnspecified, ResourceSlot
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.models.keypair import keypairs
from ai.backend.manager.models.resource_policy import (
    KeyPairResourcePolicyRow,
    ProjectResourcePolicyRow,
    UserResourcePolicyRow,
    keypair_resource_policies,
    project_resource_policies,
    user_resource_policies,
)
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.types import OptionalState, TriState

from .base import (
    BigInt,
    batch_result,
)

if TYPE_CHECKING:
    from ai.backend.manager.repositories.keypair_resource_policy.creators import (
        KeyPairResourcePolicyCreatorSpec,
    )
    from ai.backend.manager.repositories.keypair_resource_policy.updaters import (
        KeyPairResourcePolicyUpdaterSpec,
    )
    from ai.backend.manager.repositories.project_resource_policy.creators import (
        ProjectResourcePolicyCreatorSpec,
    )
    from ai.backend.manager.repositories.project_resource_policy.updaters import (
        ProjectResourcePolicyUpdaterSpec,
    )
    from ai.backend.manager.repositories.user_resource_policy.creators import (
        UserResourcePolicyCreatorSpec,
    )
    from ai.backend.manager.repositories.user_resource_policy.updaters import (
        UserResourcePolicyUpdaterSpec,
    )

    from .schema import GraphQueryContext


def _get_keypair_resource_policy_creator_spec() -> type[KeyPairResourcePolicyCreatorSpec]:
    from ai.backend.manager.repositories.keypair_resource_policy.creators import (
        KeyPairResourcePolicyCreatorSpec,
    )

    return KeyPairResourcePolicyCreatorSpec


def _get_keypair_resource_policy_updater_spec() -> type[KeyPairResourcePolicyUpdaterSpec]:
    from ai.backend.manager.repositories.keypair_resource_policy.updaters import (
        KeyPairResourcePolicyUpdaterSpec,
    )

    return KeyPairResourcePolicyUpdaterSpec


def _get_project_resource_policy_creator_spec() -> type[ProjectResourcePolicyCreatorSpec]:
    from ai.backend.manager.repositories.project_resource_policy.creators import (
        ProjectResourcePolicyCreatorSpec,
    )

    return ProjectResourcePolicyCreatorSpec


def _get_project_resource_policy_updater_spec() -> type[ProjectResourcePolicyUpdaterSpec]:
    from ai.backend.manager.repositories.project_resource_policy.updaters import (
        ProjectResourcePolicyUpdaterSpec,
    )

    return ProjectResourcePolicyUpdaterSpec


def _get_user_resource_policy_creator_spec() -> type[UserResourcePolicyCreatorSpec]:
    from ai.backend.manager.repositories.user_resource_policy.creators import (
        UserResourcePolicyCreatorSpec,
    )

    return UserResourcePolicyCreatorSpec


def _get_user_resource_policy_updater_spec() -> type[UserResourcePolicyUpdaterSpec]:
    from ai.backend.manager.repositories.user_resource_policy.updaters import (
        UserResourcePolicyUpdaterSpec,
    )

    return UserResourcePolicyUpdaterSpec


log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = (
    "CreateKeyPairResourcePolicy",
    "CreateProjectResourcePolicy",
    "CreateUserResourcePolicy",
    "DeleteKeyPairResourcePolicy",
    "DeleteProjectResourcePolicy",
    "DeleteUserResourcePolicy",
    "KeyPairResourcePolicy",
    "ModifyKeyPairResourcePolicy",
    "ModifyProjectResourcePolicy",
    "ModifyUserResourcePolicy",
    "ProjectResourcePolicy",
    "UserResourcePolicy",
)


class KeyPairResourcePolicy(graphene.ObjectType):  # type: ignore[misc]
    name = graphene.String()
    created_at = GQLDateTime()
    default_for_unspecified = graphene.String()
    total_resource_slots = graphene.JSONString()
    max_session_lifetime = graphene.Int()
    max_concurrent_sessions = graphene.Int()
    max_containers_per_session = graphene.Int()
    idle_timeout = BigInt()
    allowed_vfolder_hosts = graphene.JSONString()

    max_vfolder_count = graphene.Int(deprecation_reason="Deprecated since 23.09.4.")
    max_vfolder_size = BigInt(deprecation_reason="Deprecated since 23.09.4.")
    max_quota_scope_size = BigInt(deprecation_reason="Deprecated since 23.09.6.")
    max_concurrent_sftp_sessions = graphene.Int(description="Added in 23.03.3.")
    max_pending_session_count = graphene.Int(description="Added in 24.03.4.")
    max_pending_session_resource_slots = graphene.JSONString(description="Added in 24.03.4.")

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: Row[Any] | None,
    ) -> KeyPairResourcePolicy | None:
        if row is None:
            return None

        if row.max_pending_session_resource_slots is not None:
            max_pending_session_resource_slots = row.max_pending_session_resource_slots.to_json()
        else:
            max_pending_session_resource_slots = None
        return cls(
            name=row.name,
            created_at=row.created_at,
            default_for_unspecified=row.default_for_unspecified.name,
            total_resource_slots=row.total_resource_slots.to_json(),
            max_session_lifetime=row.max_session_lifetime,
            max_concurrent_sessions=row.max_concurrent_sessions,
            max_concurrent_sftp_sessions=row.max_concurrent_sftp_sessions,
            max_containers_per_session=row.max_containers_per_session,
            idle_timeout=row.idle_timeout,
            allowed_vfolder_hosts=row.allowed_vfolder_hosts.to_json(),
            max_pending_session_count=row.max_pending_session_count,
            max_pending_session_resource_slots=max_pending_session_resource_slots,
        )

    @classmethod
    async def load_all(cls, ctx: GraphQueryContext) -> Sequence[KeyPairResourcePolicy]:
        query = sa.select(keypair_resource_policies).select_from(keypair_resource_policies)
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def load_all_user(
        cls,
        ctx: GraphQueryContext,
        access_key: str,
    ) -> Sequence[KeyPairResourcePolicy]:
        j = sa.join(
            keypairs,
            keypair_resource_policies,
            keypairs.c.resource_policy == keypair_resource_policies.c.name,
        )
        query = (
            sa.select(keypair_resource_policies)
            .select_from(j)
            .where(
                keypairs.c.user_id
                == (
                    sa.select(keypairs.c.user_id)
                    .select_from(keypairs)
                    .where(keypairs.c.access_key == access_key)
                    .as_scalar()
                ),
            )
        )
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[KeyPairResourcePolicy | None]:
        query = (
            sa.select(keypair_resource_policies)
            .select_from(keypair_resource_policies)
            .where(keypair_resource_policies.c.name.in_(names))
            .order_by(keypair_resource_policies.c.name)
        )
        async with ctx.db.begin_readonly() as conn:
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                names,
                lambda row: row.name,
            )

    @classmethod
    async def batch_load_by_name_user(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[KeyPairResourcePolicy | None]:
        access_key = ctx.access_key
        j = sa.join(
            keypairs,
            keypair_resource_policies,
            keypairs.c.resource_policy == keypair_resource_policies.c.name,
        )
        query = (
            sa.select(keypair_resource_policies)
            .select_from(j)
            .where(
                (keypair_resource_policies.c.name.in_(names))
                & (keypairs.c.access_key == access_key),
            )
            .order_by(keypair_resource_policies.c.name)
        )
        async with ctx.db.begin_readonly() as conn:
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                names,
                lambda row: row.name,
            )

    @classmethod
    async def batch_load_by_ak(
        cls,
        ctx: GraphQueryContext,
        access_keys: Sequence[str],
    ) -> Sequence[KeyPairResourcePolicy]:
        j = sa.join(
            keypairs,
            keypair_resource_policies,
            keypairs.c.resource_policy == keypair_resource_policies.c.name,
        )
        query = (
            sa.select(keypair_resource_policies)
            .select_from(j)
            .where(keypairs.c.access_key.in_(access_keys))
            .order_by(keypair_resource_policies.c.name)
        )
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]


class CreateKeyPairResourcePolicyInput(graphene.InputObjectType):  # type: ignore[misc]
    default_for_unspecified = graphene.String(required=True)
    total_resource_slots = graphene.JSONString(required=False, default_value={})
    max_session_lifetime = graphene.Int(required=False, default_value=0)
    max_concurrent_sessions = graphene.Int(required=True)
    max_concurrent_sftp_sessions = graphene.Int(required=False, default_value=1)
    max_containers_per_session = graphene.Int(required=True)
    idle_timeout = BigInt(required=True)
    allowed_vfolder_hosts = graphene.JSONString(required=False)
    max_vfolder_count = graphene.Int(required=False, deprecation_reason="Deprecated since 23.09.4.")
    max_vfolder_size = BigInt(required=False, deprecation_reason="Deprecated since 23.09.4.")
    max_quota_scope_size = BigInt(required=False, deprecation_reason="Deprecated since 23.09.6.")
    max_pending_session_count = graphene.Int(description="Added in 24.03.4.")
    max_pending_session_resource_slots = graphene.JSONString(description="Added in 24.03.4.")

    def to_creator(self, name: str) -> Creator[KeyPairResourcePolicyRow]:
        default_for_unspecified = DefaultForUnspecified[self.default_for_unspecified]
        total_resource_slots = ResourceSlot.from_user_input(self.total_resource_slots, None)

        max_pending_session_resource_slots = (
            ResourceSlot.from_user_input(self.max_pending_session_resource_slots, None)
            if self.max_pending_session_resource_slots
            else None
        )

        def value_or_none(value: Any) -> Any:
            return value if value is not Undefined else None

        CreatorSpec = _get_keypair_resource_policy_creator_spec()
        return Creator(
            spec=CreatorSpec(
                name=name,
                default_for_unspecified=default_for_unspecified,
                total_resource_slots=total_resource_slots,
                max_session_lifetime=value_or_none(self.max_session_lifetime),
                max_concurrent_sessions=value_or_none(self.max_concurrent_sessions),
                max_concurrent_sftp_sessions=value_or_none(self.max_concurrent_sftp_sessions),
                max_containers_per_session=value_or_none(self.max_containers_per_session),
                idle_timeout=value_or_none(self.idle_timeout),
                allowed_vfolder_hosts=value_or_none(self.allowed_vfolder_hosts),
                max_vfolder_count=value_or_none(self.max_vfolder_count),
                max_vfolder_size=value_or_none(self.max_vfolder_size),
                max_quota_scope_size=value_or_none(self.max_quota_scope_size),
                max_pending_session_count=value_or_none(self.max_pending_session_count),
                max_pending_session_resource_slots=value_or_none(
                    max_pending_session_resource_slots
                ),
            )
        )


class ModifyKeyPairResourcePolicyInput(graphene.InputObjectType):  # type: ignore[misc]
    default_for_unspecified = graphene.String(required=False)
    total_resource_slots = graphene.JSONString(required=False)
    max_session_lifetime = graphene.Int(required=False)
    max_concurrent_sessions = graphene.Int(required=False)
    max_concurrent_sftp_sessions = graphene.Int(required=False)
    max_containers_per_session = graphene.Int(required=False)
    idle_timeout = BigInt(required=False)
    allowed_vfolder_hosts = graphene.JSONString(required=False)
    max_vfolder_count = graphene.Int(required=False, deprecation_reason="Deprecated since 23.09.4.")
    max_vfolder_size = BigInt(required=False, deprecation_reason="Deprecated since 23.09.4.")
    max_quota_scope_size = BigInt(required=False, deprecation_reason="Deprecated since 23.09.6.")
    max_pending_session_count = graphene.Int(description="Added in 24.03.4.")
    max_pending_session_resource_slots = graphene.JSONString(description="Added in 24.03.4.")

    def to_updater(self, name: str) -> Updater[KeyPairResourcePolicyRow]:
        default_for_unspecified = (
            DefaultForUnspecified[self.default_for_unspecified]
            if self.default_for_unspecified is not Undefined
            else Undefined
        )

        total_resource_slots = (
            ResourceSlot.from_user_input(self.total_resource_slots, None)
            if self.total_resource_slots is not Undefined
            else Undefined
        )

        UpdaterSpec = _get_keypair_resource_policy_updater_spec()
        return Updater(
            spec=UpdaterSpec(
                default_for_unspecified=OptionalState[DefaultForUnspecified].from_graphql(
                    default_for_unspecified
                ),
                total_resource_slots=OptionalState[ResourceSlot].from_graphql(total_resource_slots),
                max_session_lifetime=OptionalState[int].from_graphql(self.max_session_lifetime),
                max_concurrent_sessions=OptionalState[int].from_graphql(
                    self.max_concurrent_sessions
                ),
                max_concurrent_sftp_sessions=OptionalState[int].from_graphql(
                    self.max_concurrent_sftp_sessions
                ),
                max_containers_per_session=OptionalState[int].from_graphql(
                    self.max_containers_per_session
                ),
                idle_timeout=OptionalState[int].from_graphql(self.idle_timeout),
                allowed_vfolder_hosts=OptionalState[dict[str, Any]].from_graphql(
                    self.allowed_vfolder_hosts
                ),
                max_vfolder_count=OptionalState[int].from_graphql(self.max_vfolder_count),
                max_vfolder_size=OptionalState[int].from_graphql(self.max_vfolder_size),
                max_quota_scope_size=OptionalState[int].from_graphql(self.max_quota_scope_size),
                max_pending_session_count=TriState[int].from_graphql(
                    self.max_pending_session_count
                ),
                max_pending_session_resource_slots=TriState[dict[str, Any]].from_graphql(
                    self.max_pending_session_resource_slots
                ),
            ),
            pk_value=name,
        )


class CreateKeyPairResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = CreateKeyPairResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    resource_policy = graphene.Field(lambda: KeyPairResourcePolicy, required=False)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
        props: CreateKeyPairResourcePolicyInput,
    ) -> CreateKeyPairResourcePolicy:
        from ai.backend.manager.services.keypair_resource_policy.actions.create_keypair_resource_policy import (
            CreateKeyPairResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await graph_ctx.processors.keypair_resource_policy.create_keypair_resource_policy.wait_for_complete(
            CreateKeyPairResourcePolicyAction(props.to_creator(name))
        )

        return CreateKeyPairResourcePolicy(
            ok=True,
            msg="",
        )


class ModifyKeyPairResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = ModifyKeyPairResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
        props: ModifyKeyPairResourcePolicyInput,
    ) -> ModifyKeyPairResourcePolicy:
        from ai.backend.manager.services.keypair_resource_policy.actions.modify_keypair_resource_policy import (
            ModifyKeyPairResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await graph_ctx.processors.keypair_resource_policy.modify_keypair_resource_policy.wait_for_complete(
            ModifyKeyPairResourcePolicyAction(name, props.to_updater(name))
        )

        return ModifyKeyPairResourcePolicy(
            ok=True,
            msg="",
        )


class DeleteKeyPairResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
    ) -> DeleteKeyPairResourcePolicy:
        from ai.backend.manager.services.keypair_resource_policy.actions.delete_keypair_resource_policy import (
            DeleteKeyPairResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await graph_ctx.processors.keypair_resource_policy.delete_keypair_resource_policy.wait_for_complete(
            DeleteKeyPairResourcePolicyAction(name)
        )
        return DeleteKeyPairResourcePolicy(
            ok=True,
            msg="",
        )


class UserResourcePolicy(graphene.ObjectType):  # type: ignore[misc]
    id = graphene.ID(required=True)
    name = graphene.String(required=True)
    created_at = GQLDateTime(required=True)
    max_vfolder_count = graphene.Int(
        description="Added in 24.03.1 and 23.09.6. Limitation of the number of user vfolders."
    )  # Added in (24.03.1, 23.09.6)
    max_quota_scope_size = BigInt(
        description="Added in 24.03.1 and 23.09.2. Limitation of the quota size of user vfolders."
    )  # Added in (24.03.1, 23.09.2)
    max_vfolder_size = BigInt(deprecation_reason="Deprecated since 23.09.2.")
    max_session_count_per_model_session = graphene.Int(
        description="Added in 24.03.1 and 23.09.10. Maximum available number of sessions per single model service which the user is in charge of."
    )
    max_customized_image_count = graphene.Int(
        description="Added in 24.03.0. Maximum available number of customized images one can publish to."
    )

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: UserResourcePolicyRow | None,
    ) -> UserResourcePolicy | None:
        if row is None:
            return None
        return cls(
            id=f"UserResourcePolicy:{row.name}",
            name=row.name,
            created_at=row.created_at,
            max_vfolder_count=row.max_vfolder_count,
            max_quota_scope_size=row.max_quota_scope_size,
            max_session_count_per_model_session=row.max_session_count_per_model_session,
            max_customized_image_count=row.max_customized_image_count,
        )

    @classmethod
    async def load_all(cls, ctx: GraphQueryContext) -> Sequence[UserResourcePolicy]:
        query = sa.select(UserResourcePolicyRow)
        async with ctx.db.begin_readonly_session() as sess:
            return [
                obj
                async for r in (await sess.stream_scalars(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[UserResourcePolicy | None]:
        query = (
            sa.select(UserResourcePolicyRow)
            .where(user_resource_policies.c.name.in_(names))
            .order_by(user_resource_policies.c.name)
        )
        async with ctx.db.begin_readonly_session() as sess:
            return await batch_result(
                ctx,
                sess,
                query,
                cls,
                names,
                lambda row: row.name,
            )

    @classmethod
    async def batch_load_by_user(
        cls,
        ctx: GraphQueryContext,
        user_uuids: Sequence[uuid.UUID],
    ) -> Sequence[UserResourcePolicy]:
        from ai.backend.manager.models.user import UserRow

        query = (
            sa.select(UserRow)
            .where(UserRow.uuid.in_(user_uuids))
            .options(selectinload(UserRow.resource_policy_row))
            .order_by(UserRow.resource_policy)
        )
        async with ctx.db.begin_readonly_session() as sess:
            return [
                obj
                async for r in (await sess.stream_scalars(query))
                if (obj := cls.from_row(ctx, r.resource_policy_row)) is not None
            ]


class CreateUserResourcePolicyInput(graphene.InputObjectType):  # type: ignore[misc]
    max_vfolder_count = graphene.Int(
        description="Added in 24.03.1 and 23.09.6. Limitation of the number of user vfolders."
    )  # Added in (24.03.1, 23.09.6)
    max_quota_scope_size = BigInt(
        description="Added in 24.03.1 and 23.09.2. Limitation of the quota size of user vfolders."
    )  # Added in (24.03.1, 23.09.2)
    max_session_count_per_model_session = graphene.Int(
        description="Added in 24.03.1 and 23.09.10. Maximum available number of sessions per single model service which the user is in charge of."
    )
    max_vfolder_size = BigInt(deprecation_reason="Deprecated since 23.09.2.")
    max_customized_image_count = graphene.Int(
        description="Added in 24.03.0. Maximum available number of customized images one can publish to."
    )

    def to_creator(self, name: str) -> Creator[UserResourcePolicyRow]:
        def value_or_default(value: Any, default: int) -> int:
            return value if value is not Undefined and value is not None else default

        CreatorSpec = _get_user_resource_policy_creator_spec()
        return Creator(
            spec=CreatorSpec(
                name=name,
                max_vfolder_count=value_or_default(self.max_vfolder_count, 0),
                max_quota_scope_size=value_or_default(self.max_quota_scope_size, 0),
                max_session_count_per_model_session=value_or_default(
                    self.max_session_count_per_model_session, 0
                ),
                max_customized_image_count=value_or_default(self.max_customized_image_count, 3),
            )
        )


class ModifyUserResourcePolicyInput(graphene.InputObjectType):  # type: ignore[misc]
    max_vfolder_count = graphene.Int(
        description="Added in 24.03.1 and 23.09.6. Limitation of the number of user vfolders."
    )  # Added in (24.03.1, 23.09.6)
    max_quota_scope_size = BigInt(
        description="Added in 24.03.1 and 23.09.2. Limitation of the quota size of user vfolders."
    )  # Added in (24.03.1, 23.09.2)
    max_session_count_per_model_session = graphene.Int(
        description="Added in 24.03.1 and 23.09.10. Maximum available number of sessions per single model service which the user is in charge of."
    )
    max_customized_image_count = graphene.Int(
        description="Added in 24.03.0. Maximum available number of customized images one can publish to."
    )

    def to_updater(self, name: str) -> Updater[UserResourcePolicyRow]:
        UpdaterSpec = _get_user_resource_policy_updater_spec()
        return Updater(
            spec=UpdaterSpec(
                max_vfolder_count=OptionalState[int].from_graphql(self.max_vfolder_count),
                max_quota_scope_size=OptionalState[int].from_graphql(self.max_quota_scope_size),
                max_session_count_per_model_session=OptionalState[int].from_graphql(
                    self.max_session_count_per_model_session
                ),
                max_customized_image_count=OptionalState[int].from_graphql(
                    self.max_customized_image_count
                ),
            ),
            pk_value=name,
        )


class CreateUserResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = CreateUserResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    resource_policy = graphene.Field(lambda: UserResourcePolicy, required=False)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
        props: CreateUserResourcePolicyInput,
    ) -> CreateUserResourcePolicy:
        from ai.backend.manager.services.user_resource_policy.actions.create_user_resource_policy import (
            CreateUserResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await (
            graph_ctx.processors.user_resource_policy.create_user_resource_policy.wait_for_complete(
                CreateUserResourcePolicyAction(props.to_creator(name))
            )
        )

        return CreateUserResourcePolicy(
            ok=True,
            msg="",
        )


class ModifyUserResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = ModifyUserResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
        props: ModifyUserResourcePolicyInput,
    ) -> ModifyUserResourcePolicy:
        from ai.backend.manager.services.user_resource_policy.actions.modify_user_resource_policy import (
            ModifyUserResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await (
            graph_ctx.processors.user_resource_policy.modify_user_resource_policy.wait_for_complete(
                ModifyUserResourcePolicyAction(name, props.to_updater(name))
            )
        )

        return ModifyUserResourcePolicy(
            ok=True,
            msg="",
        )


class DeleteUserResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
    ) -> DeleteUserResourcePolicy:
        from ai.backend.manager.services.user_resource_policy.actions.delete_user_resource_policy import (
            DeleteUserResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await (
            graph_ctx.processors.user_resource_policy.delete_user_resource_policy.wait_for_complete(
                DeleteUserResourcePolicyAction(name)
            )
        )

        return DeleteUserResourcePolicy(
            ok=True,
            msg="",
        )


class ProjectResourcePolicy(graphene.ObjectType):  # type: ignore[misc]
    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
    )

    id = graphene.ID(required=True)
    name = graphene.String(required=True)
    created_at = GQLDateTime(required=True)
    max_vfolder_count = graphene.Int(
        description="Added in 24.03.1 and 23.09.6. Limitation of the number of project vfolders."
    )  #  Added in (24.03.1, 23.09.6)
    max_quota_scope_size = BigInt(
        description="Added in 24.03.1 and 23.09.2. Limitation of the quota size of project vfolders."
    )  #  Added in (24.03.1, 23.09.2)
    max_vfolder_size = BigInt(deprecation_reason="Deprecated since 23.09.2.")
    max_network_count = graphene.Int(
        description="Added in 24.12.0. Limitation of the number of networks created on behalf of project."
    )

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: ProjectResourcePolicyRow | None,
    ) -> ProjectResourcePolicy | None:
        if row is None:
            return None
        return cls(
            id=f"ProjectResourcePolicy:{row.name}",
            name=row.name,
            created_at=row.created_at,
            max_vfolder_count=row.max_vfolder_count,
            max_vfolder_size=row.max_quota_scope_size,
            max_quota_scope_size=row.max_quota_scope_size,
            max_network_count=row.max_network_count,
        )

    @classmethod
    async def load_all(cls, ctx: GraphQueryContext) -> Sequence[ProjectResourcePolicy]:
        query = sa.select(ProjectResourcePolicyRow)
        async with ctx.db.begin_readonly_session() as sess:
            return [
                obj
                async for r in (await sess.stream_scalars(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[ProjectResourcePolicy | None]:
        query = (
            sa.select(ProjectResourcePolicyRow)
            .where(project_resource_policies.c.name.in_(names))
            .order_by(project_resource_policies.c.name)
        )
        async with ctx.db.begin_readonly_session() as sess:
            return await batch_result(
                ctx,
                sess,
                query,
                cls,
                names,
                lambda row: row.name,
            )

    @classmethod
    async def batch_load_by_project(
        cls,
        ctx: GraphQueryContext,
        project_uuids: Sequence[uuid.UUID],
    ) -> Sequence[ProjectResourcePolicy]:
        from ai.backend.manager.models.group import GroupRow

        query = (
            sa.select(GroupRow)
            .where(GroupRow.id.in_(project_uuids))
            .order_by(GroupRow.resource_policy)
            .options(selectinload(GroupRow.resource_policy_row))
        )
        async with ctx.db.begin_readonly_session() as sess:
            return [
                obj
                async for r in (await sess.stream(query))
                if (obj := cls.from_row(ctx, r.resource_policy_row)) is not None
            ]


class CreateProjectResourcePolicyInput(graphene.InputObjectType):  # type: ignore[misc]
    max_vfolder_count = graphene.Int(
        description="Added in 24.03.1 and 23.09.6. Limitation of the number of project vfolders."
    )  #  Added in (24.03.1, 23.09.6)
    max_quota_scope_size = BigInt(
        description="Added in 24.03.1 and 23.09.2. Limitation of the quota size of project vfolders."
    )  #  Added in (24.03.1, 23.09.2)
    max_vfolder_size = BigInt(deprecation_reason="Deprecated since 23.09.2.")
    max_network_count = graphene.Int(
        description="Added in 24.12.0. Limitation of the number of networks created on behalf of project. Set as -1 to allow creating unlimited networks."
    )

    def to_creator(self, name: str) -> Creator[ProjectResourcePolicyRow]:
        def value_or_default(value: Any, default: int) -> int:
            return value if value is not Undefined and value is not None else default

        CreatorSpec = _get_project_resource_policy_creator_spec()
        return Creator(
            spec=CreatorSpec(
                name=name,
                max_vfolder_count=value_or_default(self.max_vfolder_count, 0),
                max_quota_scope_size=value_or_default(self.max_quota_scope_size, 0),
                max_network_count=value_or_default(self.max_network_count, 0),
            )
        )


class ModifyProjectResourcePolicyInput(graphene.InputObjectType):  # type: ignore[misc]
    max_vfolder_count = graphene.Int(
        description="Added in 24.03.1 and 23.09.6. Limitation of the number of project vfolders."
    )  #  Added in (24.03.1, 23.09.6)
    max_quota_scope_size = BigInt(
        description="Added in 24.03.1 and 23.09.2. Limitation of the quota size of project vfolders."
    )  #  Added in (24.03.1, 23.09.2)
    max_vfolder_size = BigInt(deprecation_reason="Deprecated since 23.09.2.")
    max_network_count = graphene.Int(
        description="Added in 24.12.0. Limitation of the number of networks created on behalf of project. Set as -1 to allow creating unlimited networks."
    )

    def to_updater(self, name: str) -> Updater[ProjectResourcePolicyRow]:
        UpdaterSpec = _get_project_resource_policy_updater_spec()
        return Updater(
            spec=UpdaterSpec(
                max_vfolder_count=OptionalState[int].from_graphql(self.max_vfolder_count),
                max_quota_scope_size=OptionalState[int].from_graphql(self.max_quota_scope_size),
                max_vfolder_size=OptionalState[int].from_graphql(self.max_vfolder_size),
                max_network_count=OptionalState[int].from_graphql(self.max_network_count),
            ),
            pk_value=name,
        )


class CreateProjectResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = CreateProjectResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    resource_policy = graphene.Field(lambda: ProjectResourcePolicy, required=False)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
        props: CreateProjectResourcePolicyInput,
    ) -> CreateProjectResourcePolicy:
        from ai.backend.manager.services.project_resource_policy.actions.create_project_resource_policy import (
            CreateProjectResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await graph_ctx.processors.project_resource_policy.create_project_resource_policy.wait_for_complete(
            CreateProjectResourcePolicyAction(props.to_creator(name))
        )

        return CreateProjectResourcePolicy(
            ok=True,
            msg="",
        )


class ModifyProjectResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = ModifyProjectResourcePolicyInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
        props: ModifyProjectResourcePolicyInput,
    ) -> ModifyProjectResourcePolicy:
        from ai.backend.manager.services.project_resource_policy.actions.modify_project_resource_policy import (
            ModifyProjectResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await graph_ctx.processors.project_resource_policy.modify_project_resource_policy.wait_for_complete(
            ModifyProjectResourcePolicyAction(name, props.to_updater(name))
        )

        return ModifyProjectResourcePolicy(
            ok=True,
            msg="",
        )


class DeleteProjectResourcePolicy(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
    ) -> DeleteProjectResourcePolicy:
        from ai.backend.manager.services.project_resource_policy.actions.delete_project_resource_policy import (
            DeleteProjectResourcePolicyAction,
        )

        graph_ctx: GraphQueryContext = info.context
        await graph_ctx.processors.project_resource_policy.delete_project_resource_policy.wait_for_complete(
            DeleteProjectResourcePolicyAction(name)
        )

        return DeleteProjectResourcePolicy(
            ok=True,
            msg="",
        )



from ai.backend.common.clients.prometheus.client import PrometheusClient
from ai.backend.manager.repositories.metric.repository import MetricRepository

from .container_metric import ContainerUtilizationMetricService


class UtilizationMetricService:
    container: ContainerUtilizationMetricService
    _metric_repository: MetricRepository

    def __init__(
        self,
        prometheus_client: PrometheusClient,
        timewindow: str,
        metric_repository: MetricRepository,
    ) -> None:
        self.container = ContainerUtilizationMetricService(prometheus_client, timewindow=timewindow)
        self._metric_repository = metric_repository



from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class AppProxyConnectionError(BackendAIError, web.HTTPServiceUnavailable):
    """Raised when connection to AppProxy fails."""

    error_type = "https://api.backend.ai/probs/appproxy-connection-error"
    error_title = "Failed to connect to AppProxy."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.APPPROXY,
            operation=ErrorOperation.REQUEST,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class AppProxyResponseError(BackendAIError, web.HTTPInternalServerError):
    """Raised when AppProxy returns an invalid response."""

    error_type = "https://api.backend.ai/probs/appproxy-response-error"
    error_title = "Invalid response from AppProxy."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.APPPROXY,
            operation=ErrorOperation.REQUEST,
            error_detail=ErrorDetail.INVALID_DATA_FORMAT,
        )



import logging

from ai.backend.common.leader.tasks import EventTaskSpec
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.sokovan.deployment.coordinator import DeploymentCoordinator
from ai.backend.manager.sokovan.deployment.route.coordinator import RouteCoordinator
from ai.backend.manager.sokovan.scheduler.coordinator import ScheduleCoordinator

log = BraceStyleAdapter(logging.getLogger(__name__))


class SokovanOrchestrator:
    """
    Orchestrator for Sokovan scheduler.
    Provides event handler interface for the coordinator.
    """

    _deployment_coordinator: DeploymentCoordinator
    _route_coordinator: RouteCoordinator
    _schedule_coordinator: ScheduleCoordinator

    def __init__(
        self,
        schedule_coordinator: ScheduleCoordinator,
        deployment_coordinator: DeploymentCoordinator,
        route_coordinator: RouteCoordinator,
    ) -> None:
        self._schedule_coordinator = schedule_coordinator
        self._deployment_coordinator = deployment_coordinator
        self._route_coordinator = route_coordinator

    @property
    def coordinator(self) -> ScheduleCoordinator:
        """Get the schedule coordinator."""
        return self._schedule_coordinator

    @property
    def deployment_coordinator(self) -> DeploymentCoordinator:
        """Get the deployment coordinator."""
        return self._deployment_coordinator

    @property
    def route_coordinator(self) -> RouteCoordinator:
        """Get the route coordinator."""
        return self._route_coordinator

    def create_task_specs(self) -> list[EventTaskSpec]:
        """Create task specifications for leader-based scheduling."""
        specs: list[EventTaskSpec] = []

        # Collect task specs from all coordinators
        specs.extend(self._schedule_coordinator.create_task_specs())
        specs.extend(self._deployment_coordinator.create_task_specs())
        specs.extend(self._route_coordinator.create_task_specs())

        return specs



"""Result types for scheduling operations."""

from __future__ import annotations

from dataclasses import dataclass

from ai.backend.common.types import AccessKey, SessionId


@dataclass
class ScheduledSessionData:
    """Data for a scheduled session."""

    session_id: SessionId
    creation_id: str
    access_key: AccessKey
    reason: str
